"""

Examples of table definitions

``` python
from nagra import Table

city = Table(
    "city",
    columns={
        "name": "varchar",
        "lat": "varchar",
        "long": "varchar",
    },
    natural_key=["name"],
    one2many = {
        "temperatures": "temperature.city",
    }
)

temperature = Table(
    "temperature",
    columns={
        "timestamp": "timestamp",
        "city": "int",
        "value": "float",
    },
    natural_key=["city", "timestamp"],
    foreign_keys={
        "city": "city",
    },

)
```


"""

import warnings
from datetime import date, datetime
from functools import lru_cache
from typing import Iterable, Optional, Union, TYPE_CHECKING

from nagra.delete import Delete
from nagra.exceptions import IncorrectSchema
from nagra.schema import Schema
from nagra.select import Select
from nagra.sexpr import AST
from nagra.statement import Statement
from nagra.transaction import Transaction
from nagra.update import Update
from nagra.copy import copy_from
from nagra.upsert import Upsert
from nagra.utils import quote_identifier

if TYPE_CHECKING:
    from pandas import DataFrame


# Intentionally sorted by reverse length to help type hint detection,
# see Schema._db_columns
_TYPE_ALIAS = {
    "timestamp without time zone": "timestamp",
    "timestamp with time zone": "timestamptz",
    "character varying": "str",
    "double precision": "float",
    "timestamptz": "timestamptz",
    "timestamp": "timestamp",
    "varbinary": "blob",
    "datetime": "timestamp",
    "nvarchar": "str",
    "smallint": "int",
    "boolean": "bool",
    "decimal": "float",
    "integer": "int",
    "numeric": "float",
    "tinyint": "int",
    "varchar": "str",
    "bigint": "bigint",
    "vector": "float []",
    "bytea": "blob",
    "bytes": "blob",
    "float": "float",
    "jsonb": "json",
    "money": "float",
    "nchar": "str",
    "char": "str",
    "blob": "blob",
    "cidr": "str",
    "inet": "str",
    "real": "float",
    "bool": "bool",
    "date": "date",
    "json": "json",
    "text": "str",
    "uuid": "uuid",
    "int": "int",
    "bit": "bool",
    "uniqueidentifier": "uuid",
    "datetime2": "timestamp",
    "smalldatetime": "timestamp",
    "datetimeoffset": "timestamptz",
    "str": "str",
    "": "str",
}

_DB_TYPE = {
    "postgresql": {
        "str": "TEXT",
        "int": "INTEGER",
        "bigint": "BIGINT",
        "float": "FLOAT",
        "timestamp": "TIMESTAMP",
        "timestamptz": "TIMESTAMPTZ",
        "date": "DATE",
        "bool": "BOOL",
        "uuid": "UUID",
        "json": "JSON",
        "blob": "BYTEA",
    },
    "sqlite": {
        "str": "TEXT",
        "int": "INTEGER",
        "bigint": "INTEGER",
        "float": "FLOAT",
        "timestamp": "DATETIME",
        "timestamptz": "DATETIME",
        "date": "DATE",
        "bool": "BOOL",
        "uuid": "TEXT",
        "json": "JSON",
        "blob": "BLOB",
    },
    "mssql": {
        "str": "NVARCHAR(200)",  # Could use MAX here but then it is not usable as index
        "int": "INT",
        "bigint": "BIGINT",
        "float": "FLOAT",
        "timestamp": "DATETIME2",
        "timestamptz": "DATETIMEOFFSET",
        "date": "DATE",
        "bool": "BIT",
        "uuid": "UNIQUEIDENTIFIER",
        "json": "NVARCHAR(MAX)",
        "blob": "VARBINARY(MAX)",
    },
}


class Column:
    __slots__ = ["name", "dtype", "dims"]

    def __init__(self, name: str, dtype: str):
        self.name = name.strip()
        if "[" in dtype:
            dtype, dims = dtype.split("[", 1)
            self.dtype = dtype.strip()
            self.dims = "[" + dims.strip()
        else:
            self.dtype = dtype.strip()
            self.dims = ""
        try:
            self.dtype = _TYPE_ALIAS[dtype.strip().lower()]
        except KeyError:
            self.dtype = "str"
            warnings.warn(
                f"Type '{dtype}' not supported (for column '{name}'), falling back to string type."
            )

    def eq(self, other):
        if not isinstance(other, Column):
            return False
        return self.name == other.name and self.dtype == other.dtype

    def python_type(self):
        res = None
        match self.dtype:
            case "int" | "bigint":
                res = int
            case "str":
                res = str
            case "float":
                res = float
            case "timestamp" | "timestamptz":
                res = datetime
            case "bool":
                res = bool
            case "json":
                res = list | dict
            case "date":
                res = date
            case "uuid":
                res = str
            case "blob":
                res = bytes
            case _:
                raise RuntimeError("Unexpected error")

        for c in self.dims:
            if c == "[":
                res = list[res]

        return res


class Table:
    def __init__(
        self,
        name: str,
        columns: dict,
        natural_key: Optional[list[str]] = None,
        foreign_keys: Optional[dict[str, str]] = None,
        not_null: Optional[list[str]] = None,
        one2many: Optional[dict] = None,
        default: Optional[dict[str, str]] = None,
        primary_key: Optional[str] = "id",
        schema: Schema = Schema.default,
        is_view: Optional[bool] = False,
    ):
        self.name = name
        self.columns: dict[str, Column] = {
            name: Column(name, dtype) for name, dtype in columns.items()
        }
        self.natural_key = natural_key or []
        self.foreign_keys = foreign_keys or {}
        self.not_null = set(not_null or [])
        self.one2many = one2many or {}
        self.default = default or {}
        self.primary_key = primary_key
        self.schema = schema
        self.is_view = is_view

        # Detect malformed fk definitions
        if len(self.natural_key) == 1:
            (nk,) = self.natural_key
            for fk, fk_table in self.foreign_keys.items():
                if fk != nk or fk_table != name:
                    continue
                msg = f"Table '{name}': Foreign key '{fk}' refers to table natural key"
                raise IncorrectSchema(msg)

        # Detect incorrect nk
        for nk_name in self.natural_key:
            if nk_name not in self.columns:
                raise IncorrectSchema(
                    f"Table '{name}': unknown column name '{nk_name}'"
                    " referenced in natural key"
                )

        # natural key and primary can be both unset/empty,
        # but it may be undesirable
        if not self.primary_key and not self.natural_key:
            warnings.warn(f"Table '{name}': no primary key or natural key defined")

        # Add table to schema
        self.schema.add_table(self.name, self)

    @classmethod
    def get(self, name, schema=Schema.default) -> "Table":
        """
        Shortcut method to Schema.default().get()
        """
        return schema.tables[name]

    def select(self, *columns, distinct: bool = False, trn=None):
        trn = trn or Transaction.current()
        if not columns:
            columns = self.default_columns()
        slct = Select(self, *columns, distinct=distinct, trn=trn, env=Env(self))
        return slct

    def select_distinct(self, *columns, trn=None):
        return self.select(*columns, distinct=True, trn=trn)

    def delete(self, where=None, trn=None):
        trn = trn or Transaction.current()
        delete = Delete(self, trn=trn, env=Env(self))
        if where:
            return delete.where(where)
        return delete

    def upsert(
        self,
        *columns,
        trn: Optional[Transaction] = None,
        lenient: Union[bool, list[str]] = False,
    ):
        """
        Create an upsert object based on the given columns, if
        lenient is set, foreign keys wont be enforced on the given
        columns even if a value is passed on the subsequent execute or
        executemany. Example:

        >>> upsert = Table.get("comment").upsert("body", "blog_post.title", lenient=["blog_post"])
        >>> upsert.execute(("Nice post!", "A post title that will change soon."))

        If lenient is set to True all foreign keys will be treated as such.
        """
        if not columns:
            columns = self.default_columns()
        trn = trn or Transaction.current()
        return Upsert(self, *columns, trn=trn, env=Env(self), lenient=lenient)

    def update(
        self,
        *columns,
        trn: Optional[Transaction] = None,
        lenient: Union[bool, list[str]] = False,
    ):
        if not columns:
            columns = self.default_columns()
        trn = trn or Transaction.current()
        return Update(self, *columns, trn=trn, lenient=lenient)

    def insert(
        self,
        *columns,
        trn: Optional[Transaction] = None,
        lenient: Union[bool, list[str]] = False,
    ):
        """
        Provide an insert-only statement (won't raise error if
        record already exists). See `Table.upsert` for `lenient` role.
        """
        trn = trn or Transaction.current()
        return self.upsert(*columns, trn=trn, lenient=lenient).insert_only()

    def copy_from(
        self,
        rows: Iterable[tuple] | "DataFrame",
        trn: Optional[Transaction] = None,
        lenient: Union[bool, list[str]] = False,
    ):
        """
        Execute a COPY <table> FROM STDIN (only supported with
        postgresql). See `Table.upsert` for `lenient` role.
        """
        trn = trn or Transaction.current()
        return copy_from(self, rows=rows, trn=trn, lenient=lenient)

    def drop(self, trn: Optional[Transaction] = None):
        trn = trn or Transaction.current()
        stmt = Statement("drop_table", trn.flavor, name=self.name)
        trn.execute(stmt())

    def required(self, col_name):
        return (
            col_name in self.natural_key
            or col_name in self.not_null
            or col_name == self.primary_key
        )

    def default_columns(self, compact: bool = False, skip_pk=False, skip_blob=False):
        """
        Return the list of default column for the current
        table. Used by `Table.select` and `Table.upsert` when no
        columns are provided.
        """
        if compact:
            columns = self.natural_key or [self.primary_key]
        else:
            columns = self.columns

        for column in columns:
            # Skip pk if asked
            if skip_pk and column in self.primary_key:
                continue
            # Skip blobs
            if skip_blob and self.column[column].dtype == "blob":
                continue
            # Escape literals (nul, true, false)
            if column in AST.literals:
                yield f".{column}"
                continue
            # Handle non foreign keys
            if column not in self.foreign_keys or compact:
                yield column
                continue
            # FK
            ftable = self.schema.get(self.foreign_keys[column])
            yield from (f"{column}.{k}" for k in ftable.default_columns(compact=True))

    def join(self, env: "Env"):
        for prefix, alias in env.refs.items():
            # Find alias of previous join in the chain
            *head, tail = prefix
            prev_table = env.refs[tuple(head)] if head else self.name
            # Identify last table & column of the chain
            ftable, alias_col, join_col = self.join_on(prefix, env)
            yield (ftable.name, alias, prev_table, alias_col, join_col)

    @lru_cache
    def join_on(self, path: tuple[str, ...], env: "Env") -> tuple["Table", str, str]:
        """
        `path` is a tuple containing names of column, each of
        which is a foreign key to another table.

        Returns the next table to join and the column to join on.

        """
        if len(path) == 1:
            head = path[0]
            if alias := self.one2many.get(head):
                # An alias is a string containing "table_name.fk_name"
                # that points to current table
                table_name, alias_col = alias.split(".")
                ftable = self.schema.get(table_name)
                join_col = ftable.primary_key
            else:
                # not an alias we implictly join on self, based on the
                # given column
                join_col = head
                fname = self.foreign_keys[join_col]
                ftable = self.schema.get(fname)
                alias_col = ftable.primary_key
            return ftable, alias_col, join_col

        # Recurse to find the previous table in the chain
        prev_table, *_ = self.join_on(path[:-1], env)
        # Resolve last step
        return prev_table.join_on(path[-1:], env)

    def ctypes(self, flavor: str, column_names: Iterable[str]):
        # detect arrays
        db_type = _DB_TYPE[flavor]
        res = {}
        for name in column_names:
            col = self.columns[name]
            if flavor == "sqlite" and col.dims:
                # sqlite does not support arrays
                res[name] = "json"
            else:
                res[name] = db_type[col.dtype] + col.dims
        return res

    @property
    def has_array(self):
        """
        Return True if at least one column is an array
        """
        return any(c.dims for c in self.columns.values())

    @property
    def primary_key_is_identity(self):
        col = self.columns.get(self.primary_key)
        if not col and self.primary_key == "id":
            # TODO id type shouldn't be implicit
            return True
        return col.dtype in ("int", "bigint")

    def __iter__(self):
        return iter(self.select())

    def __repr__(self):
        return f"<Table {self.name}>"

    def eq(self, other):
        """
        Return True is `self` is equivalent to `other`, so if all
        attributes themselves are equivalent.
        """
        if id(self) == id(other):
            return True
        if not isinstance(other, Table):
            return False

        ok = all((
            self.name == other.name,
            all(a.eq(b) for a, b in zip(
                self.columns.values(), other.columns.values()
            )),
            self.primary_key == other.primary_key,
            self.natural_key == other.natural_key,
            self.foreign_keys == other.foreign_keys,
            self.not_null == other.not_null,
            self.one2many == other.one2many,
            self.default == other.default,
            self.is_view == other.is_view,
        ))
        return ok


class Env:
    def __init__(self, table: "Table", refs: Optional[dict] = None):
        self.table = table
        self.refs = refs or {}

    def add_ref(self, path, flavor):
        *head, name, tail = path
        prefix = tuple([*head, name])
        table_alias = self.refs.get(prefix)
        if not table_alias:
            if len(prefix) >= 2:
                self.add_ref(prefix, flavor)
            table_alias = f"{name}_{len(self.refs)}"
            self.refs[prefix] = table_alias
        alias = quote_identifier(table_alias, flavor)
        column = quote_identifier(tail, flavor)
        return f"{alias}.{column}"

    def __repr__(self):
        content = repr(self.refs)
        return f"<Env {self.table.name} {content}>"

    def clone(self):
        return Env(self.table, self.refs.copy())



from functools import partial

from nagra.utils import template
from nagra.transaction import dummy_transaction


class Statement:
    def __init__(self, template, flavor=None, **params):
        self._template = template
        self._params = params
        self._flavor = flavor or dummy_transaction.flavor

    def __call__(self):
        path = f"{self._flavor}/{self._template}.sql"
        return template(path).render(**self._params, _all_params=self._params)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise ValueError(f"Invalid statement parameter {name}")
        return partial(self._setter, name)

    def _setter(self, name, value):
        self._params[name] = value
        return self



from datetime import datetime
import pytest

from nagra import Transaction
from nagra.utils import strip_lines


def test_simple_select(person):
    stm = person.select("name").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT "person"."name" FROM "person" ;'


def test_select_with_join(person):
    stm = person.select("name", "parent.parent.name").stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name", "parent_1"."name"',
        'FROM "person"',
        'LEFT JOIN "person" as "parent_0" ON (',
        '"parent_0"."id" = "person"."parent"',
        ")",
        'LEFT JOIN "person" as "parent_1" ON (',
        '"parent_1"."id" = "parent_0"."parent"',
        ")",
        ";",
    ]


def test_kitchensink_select(kitchensink):
    stm = kitchensink.select().stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"kitchensink"."varchar", "kitchensink"."bigint", "kitchensink"."float", '
        '"kitchensink"."int", "kitchensink"."timestamp", "kitchensink"."timestamptz", '
        '"kitchensink"."bool", "kitchensink"."date", "kitchensink"."json", '
        '"kitchensink"."uuid", "kitchensink"."max", "kitchensink"."true", '
        '"kitchensink"."blob"',
        'FROM "kitchensink"',
        ";",
    ]

    stm = kitchensink.select(".true").stm()
    res = strip_lines(stm)
    assert res == ["SELECT", '"kitchensink"."true"', 'FROM "kitchensink"', ";"]

    stm = kitchensink.select("true").stm()
    res = strip_lines(stm)
    assert res == ["SELECT", "true", 'FROM "kitchensink"', ";"]


def test_select_clone(person):
    queries = [
        person.select("name")
        .limit(1)
        .offset(1)
        .where("(= name 'spam')")
        .groupby("name"),
        person.select("name")
        .offset(1)
        .where("(= name 'spam')")
        .groupby("name")
        .limit(1),
        person.select("name")
        .where("(= name 'spam')")
        .groupby("name")
        .limit(1)
        .offset(1),
        person.select("name")
        .groupby("name")
        .limit(1)
        .offset(1)
        .where("(= name 'spam')"),
    ]

    expected = queries[0].stm()
    for q in queries[1:]:
        assert q.stm() == expected


def test_select_with_where(person):
    select = person.select("name").where("(= id {})")
    stm = select.stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name"',
        'FROM "person"',
        "WHERE",
        '"person"."id" = %s',
        ";",
    ]

    stm = select.where("(= name 'spam')").stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name"',
        'FROM "person"',
        "WHERE",
        '"person"."id" = %s AND "person"."name" = \'spam\'',
        ";",
    ]


def test_select_where_and_join(person):
    select = person.select("name").where("(= parent.name 'foo')")
    stm = select.stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name"',
        'FROM "person"',
        'LEFT JOIN "person" as "parent_0" ON (',
        '"parent_0"."id" = "person"."parent"',
        ")",
        "WHERE",
        '"parent_0"."name" = \'foo\'',
        ";",
    ]


@pytest.mark.parametrize("op", ["min", "max", "sum"])
def test_simple_agg(person, op):
    # MIN
    stm = person.select(f"({op} name)").stm()
    res = " ".join(strip_lines(stm))
    assert res == f'SELECT {op}("person"."name") FROM "person" ;'


def test_count(person):
    stm = person.select("(count *)").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT count(*) FROM "person" ;'

    stm = person.select("(count 1)").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT count(1) FROM "person" ;'


def test_groupby(person):
    # Explicit
    stm = person.select("name", "(count *)").groupby("name").stm()
    res = " ".join(strip_lines(stm))
    assert (
        res
        == 'SELECT "person"."name", count(*) FROM "person" GROUP BY "person"."name" ;'
    )

    # implicit
    stm = person.select("name", "(count *)").stm()
    res = " ".join(strip_lines(stm))
    assert (
        res
        == 'SELECT "person"."name", count(*) FROM "person" GROUP BY "person"."name" ;'
    )


def test_orderby(person):
    # asc
    stm = person.select("name").orderby("name").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT "person"."name" FROM "person" ORDER BY "person"."name" asc ;'

    # desc
    stm = person.select("name").orderby(("name", "desc")).stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT "person"."name" FROM "person" ORDER BY "person"."name" desc ;'

    # with join
    stm = person.select("name").orderby("parent.name").stm()
    res = " ".join(strip_lines(stm))
    assert res == (
        'SELECT "person"."name" FROM "person" LEFT JOIN "person" as "parent_0" ON ( '
        '"parent_0"."id" = "person"."parent" ) ORDER BY "parent_0"."name" asc ;'
    )


def test_o2m_stm(person, org):
    # Combine one2many and implicit joins
    select = person.select(
        "name",
        "orgs.name",
        "parent.name",
        "parent.parent.name",
    )
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"person"."name", "orgs_0"."name", "parent_1"."name", "parent_2"."name"',
        'FROM "person"',
        'LEFT JOIN "org" as "orgs_0" ON (',
        '"orgs_0"."person" = "person"."id"',
        ")",
        'LEFT JOIN "person" as "parent_1" ON (',
        '"parent_1"."id" = "person"."parent"',
        ")",
        'LEFT JOIN "person" as "parent_2" ON (',
        '"parent_2"."id" = "parent_1"."parent"',
        ")",
        ";",
    ]
    assert res == expected

    # Multiple one2many
    select = person.select(
        "name",
        "orgs.country",
        "skills.name",
    )
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"person"."name", "orgs_0"."country", "skills_1"."name"',
        'FROM "person"',
        'LEFT JOIN "org" as "orgs_0" ON (',
        '"orgs_0"."person" = "person"."id"',
        ")",
        'LEFT JOIN "skill" as "skills_1" ON (',
        '"skills_1"."person" = "person"."id"',
        ")",
        ";",
    ]
    assert res == expected

    # Use a on2many after a many2one
    select = org.select(
        "name",
        "person.name",
        "person.skills.name",
    )
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"org"."name", "person_0"."name", "skills_1"."name"',
        'FROM "org"',
        'LEFT JOIN "person" as "person_0" ON (',
        '"person_0"."id" = "org"."person"',
        ")",
        'LEFT JOIN "skill" as "skills_1" ON (',
        '"skills_1"."person" = "person_0"."id"',
        ")",
        ";",
    ]
    assert res == expected

    # Use a on2many after a one2many
    select = person.select("name", "orgs.addresses.city")
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"person"."name", "addresses_1"."city"',
        'FROM "person"',
        'LEFT JOIN "org" as "orgs_0" ON (',
        '"orgs_0"."person" = "person"."id"',
        ")",
        'LEFT JOIN "address" as "addresses_1" ON (',
        '"addresses_1"."org" = "orgs_0"."id"',
        ")",
        ";",
    ]
    assert res == expected


def test_o2m_select(transaction, person, org, address):
    # Test with actual data
    person.upsert("name").execute("Charly")
    org.upsert("name", "person.name").execute("Alpha", "Charly")
    org.upsert("name", "person.name").execute("Beta", "Charly")
    address.upsert("city", "org.name").executemany(
        [
            ("Ankara", "Alpha"),
            ("Athens", "Alpha"),
            ("Beirut", "Beta"),
        ]
    )
    # if transaction.flavor == "mssql":
    #     breakpoint()
    rows = list(
        person.select("name", "orgs.addresses.city").orderby("orgs.addresses.city")
    )
    assert rows == [("Charly", "Ankara"), ("Charly", "Athens"), ("Charly", "Beirut")]


def test_agg(transaction, temperature):
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-01", "Berlin", 10),
            ("1970-01-01", "London", 12),
        ]
    )
    rows = list(
        temperature.select(
            "city",
        )
    )
    assert len(rows) == 2

    # String concat
    match Transaction.current().flavor:
        case "postgresql" | "mssql":
            select = temperature.select("(string_agg city ',')")
        case "sqlite":
            select = temperature.select("(group_concat city)")

    rows = list(select)
    assert len(rows) == 1
    (record,) = rows
    assert record[0] in ("Berlin,London", "London,Berlin")

    # Strings into array
    if Transaction.current().flavor == "postgresql":
        (record,) = list(temperature.select("(array_agg city)"))
        assert sorted(record[0]) == ["Berlin", "London"]

    # sum, avg, min and max
    for op, expected in [("sum", 22), ("min", 10), ("max", 12), ("avg", 11)]:
        select = temperature.select(f"({op} value)")
        (record,) = list(select)
        assert expected == record[0]

    # Add more rows
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )

    records = dict(temperature.select("city", "(sum value)").groupby("city"))
    assert records == {"Berlin": 20.0, "London": 24.0}

    # Json agg
    if Transaction.current().flavor == "postgresql":
        select = temperature.select("(json_object_agg city value)")
        (record,) = list(select)
        assert record[0] == {"Berlin": 10, "London": 12}


def test_date_op(transaction, temperature):
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    match Transaction.current().flavor:
        case "postgresql":
            select = temperature.select("(extract 'year' timestamp)")
            records = list(select)
            assert records[0][0] == 1970
        case "mssql":
            select = temperature.select("(year timestamp)")
            records = list(select)
            assert records[0][0] == 1970
        case "sqlite":
            select = temperature.select("(strftime '%Y' timestamp)")
            records = list(select)
            assert records[0][0] == "1970"
    assert len(records) == 2


def test_to_dict(transaction, temperature):
    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    # Read data
    expected_date = datetime(1970, 1, 2, 0, 0)
    if transaction.flavor == "sqlite":
        expected_date = str(expected_date.date())

    records = list(temperature.select().orderby("city").to_dict())
    assert len(records) == 2
    assert records[0] == {
        "timestamp": expected_date,
        "city": "Berlin",
        "value": 10.0,
    }
    # Read with custom arg
    cond = "(= value {})"
    (record,) = temperature.select().where(cond).to_dict(12)
    assert record == {
        "timestamp": expected_date,
        "city": "London",
        "value": 12.0,
    }


def test_select_alias(transaction, temperature):
    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    select = temperature.select().orderby("city").aliases("t", "c", "v")

    # Check record keys
    records = list(select.to_dict())
    assert sorted(records[0]) == ["c", "t", "v"]

    # Check pandas df columns
    df = select.to_pandas()
    assert sorted(df) == ["c", "t", "v"]

    # Check polars df columns
    if transaction.flavor == "postgresql":
        df = select.to_polars().collect()
        assert sorted(df.columns) == ["c", "t", "v"]


@pytest.mark.parametrize("nest_with_param", [False, True])
def test_to_nested_dict(transaction, person, nest_with_param):
    # Upsert
    person.upsert("name", "parent.name").execute("Alice", None)
    person.upsert("name", "parent.name").execute("Bob", "Alice")
    person.upsert("name", "parent.name").execute("Charly", "Bob")

    # Read data
    select = person.select("name", "parent.name", "parent.parent.name")
    if nest_with_param:
        records = list(select.to_dict(nest=True))
    else:
        records = list(select.to_nested_dict())

    assert records == [
        {
            "name": "Alice",
            "parent": None,
        },
        {
            "name": "Bob",
            "parent": {
                "name": "Alice",
                "parent": None,
            },
        },
        {
            "name": "Charly",
            "parent": {
                "name": "Bob",
                "parent": {
                    "name": "Alice",
                },
            },
        },
    ]


def test_any_and_values(transaction, person):
    person.insert("id", "name").execute(1, "one")
    if transaction.flavor != "mssql":
        # VALUES is not supported by mssql
        select = person.select("id").where("(in id (values {} {}))")
        res = list(select.execute(1, 2))
        assert res == [(1,)]

    if transaction.flavor == "postgresql":
        # ANY is only supported by postgres

        select = person.select("id").where("(= id (any {}))")
        res = list(select.execute([1, 2]))
        assert res == [(1,)]


def test_distinct_on(transaction, person):
    if transaction.flavor != "postgresql":
        pytest.skip("Disctinct on is only supported by PostgreSQL")

    person.insert("name").execute("one")
    person.insert("name", "parent.name").executemany(
        [
            ("two", "one"),
            ("three", "one"),
        ]
    )

    select = (
        person.select("name")
        .distinct_on("parent.name")
        .orderby(
            "parent.name",
            "name",
        )
        .where(
            "(isnot parent null)",
        )
    )
    assert list(select) == [("three",)]


def test_select_distinct(transaction, temperature):
    if transaction.flavor == "mssql":
        pytest.skip("Disctinct on is not supported by MSSQL")

    temperature.insert("city", "timestamp", "value").executemany(
        [
            ("Brussels", "2025-10-03", 11),
            ("Amsterdam", "2025-10-03", 11),
        ]
    )

    select = temperature.select_distinct(
        "timestamp",
        "value",
    )
    dt = datetime(2025, 10, 3) if transaction.flavor == "postgresql" else "2025-10-03"
    assert list(select) == [(dt, 11)]



from itertools import chain, groupby
from contextlib import contextmanager
from collections import defaultdict
from pathlib import Path
from io import IOBase
from typing import Optional, TYPE_CHECKING
from warnings import warn

import toml
from nagra.statement import Statement
from nagra.transaction import DummyTransaction, Transaction
from nagra.utils import logger, snake_to_pascal, template


if TYPE_CHECKING:
    from nagra.table import Table
    from nagra.view import View

MSSQL_ARRAY_MSG = (
    "MS SQL Server does not support array types. Table '{table}' "
    "with array columns is ignored."
)


class Schema:
    default: "Schema" = None

    def __init__(self, tables=None, views=None):
        self.tables: dict[str, Table] = tables or {}
        self.views: dict[str, View] = views or {}

    @classmethod
    def from_toml(cls, toml_src: IOBase | Path | str) -> "Schema":
        schema = cls()
        schema.load_toml(toml_src)
        return schema

    @property
    def empty(self) -> bool:
        return not self.tables

    def load_toml(self, toml_src: IOBase | Path | str):
        # Late import to avoid import loops
        from nagra.table import Table
        from nagra.view import View

        # load table definitions
        match toml_src:
            case IOBase():
                content = toml_src.read()
            case Path():
                with toml_src.open() as toml_src:
                    content = toml_src.read()
            case _:
                content = toml_src
        tables = toml.loads(content)
        # Instanciate tables
        for name, info in tables.items():
            logger.debug("Instanciate '%s' table from toml", name)
            if "primary_key" in info:
                if info["primary_key"].strip() == "":
                    info["primary_key"] = None

            # Handle view info
            view_params = ("view_columns", "as_select", "view_select")
            if any(info.get(c) for c in view_params):
                View(name, **info, schema=self)
            else:
                Table(name, **info, schema=self)

    def add_table(self, name: str, table: "Table"):
        if name in self.tables:
            raise RuntimeError(f"Table {name} already in schema!")
        self.tables[name] = table

    def add_view(self, name: str, view: "View"):
        if name in self.views:
            raise RuntimeError(f"View {name} already in schema!")
        self.views[name] = view

    def reset(self):
        self.tables = {}
        self.views = {}

    def get(self, name: str) -> "Table | View":
        """
        Return the view or the table with name `name`
        """
        # We must get() on views first, since each view is also
        # registered as a table
        res = self.views.get(name) or self.tables[name]
        if not res:
            raise KeyError(f"No view or table named {name}")
        return res

    @classmethod
    def _db_columns(
        cls, trn=None, pg_schema="public"
    ) -> defaultdict[str, dict[str, str]]:
        """
        Return a mapping of table -> column -> type by introspecting the database
        """
        from nagra.table import _TYPE_ALIAS

        trn = trn or Transaction.current()
        res = defaultdict(dict)
        stmt = Statement("find_columns", trn.flavor, pg_schema=pg_schema)
        for tbl, col_name, col_type, *hints in trn.execute(stmt()):
            # handle array types
            if col_type.upper() == "ARRAY" and hints:
                # Try to find type of elements, rely on the fact that
                # _TYPE_ALIAS keys are sorted by reverse length.
                hint = hints[0].lower()
                for candidate in _TYPE_ALIAS:
                    if candidate in hint:
                        # XXX We can not detect array dimensions,
                        # fallback to simple array
                        col_type = f"{candidate} []"
                        break
                else:
                    msg = f"Unable to detect type of column: {col_name} in table {tbl}"
                    raise RuntimeError(msg)
            # handle pg extension types
            if col_type.upper() == "USER-DEFINED" and hints:
                for name in hints:
                    if col_type := _TYPE_ALIAS.get(name.lower()):
                        break
                else:
                    msg = f"Unable to detect type of column: {col_name} in table {tbl}"
                    raise RuntimeError(msg)

            res[tbl][col_name] = col_type
        return res

    def _db_indexes(self, trn=None, pg_schema="public"):
        trn = trn or Transaction.current()
        stmt = Statement("find_indexes", trn.flavor, pg_schema=pg_schema)
        res = [n for (n,) in trn.execute(stmt())]
        return res

    def _db_views(self, trn=None, pg_schema="public") -> dict[str, str]:
        trn = trn or Transaction.current()
        stmt = Statement("find_views", trn.flavor, pg_schema=pg_schema)
        # The statement returns tuples of (name, view_def)
        res = dict(trn.execute(stmt()))
        return res

    @classmethod
    def _db_fk(
        cls, *whitelist, trn=None, pg_schema="public"
    ) -> defaultdict[str, dict[str, "FKConstraint"]]:
        trn = trn or Transaction.current()
        res = defaultdict(dict)
        stmt = Statement(
            "find_foreign_keys",
            trn.flavor,
            pg_schema=pg_schema,  # FIXME put schema on transaction
            mssql_schema="dbo",  # should come from the dsn
        )
        for name, tbl, col, ftable, fcol in trn.execute(stmt()):
            if whitelist and tbl not in whitelist:
                continue
            if name in res[tbl]:
                raise RuntimeError("Unexpected multi-columns foreign key")
            res[tbl][name] = FKConstraint(name, tbl, col, ftable, fcol)
        return res

    @classmethod
    def _db_pk(cls, trn=None, pg_schema="public"):
        trn = trn or Transaction.current()
        res = {}
        stmt = Statement("find_primary_keys", trn.flavor, pg_schema=pg_schema)
        for tbl, pk_col in trn.execute(stmt()):
            if tbl in res:
                raise RuntimeError("Unexpected multi-columns primary key")
            res[tbl] = pk_col
        return res

    @classmethod
    def _db_unique(cls, db_pk, trn=None, pg_schema="public") -> dict[str, list[str]]:
        trn = trn or Transaction.current()
        by_constraint = defaultdict(list)

        if trn.flavor == "sqlite":
            stmt = Statement("find_unique_constraint", trn.flavor)
            constraints = trn.execute(stmt()).fetchall()
            for (tbl, idx_name), rows in groupby(constraints, key=lambda x: x[:2]):
                rows = list(rows)
                by_constraint[tbl].append([col_name for _, _, col_name in rows])

        else:
            stmt = Statement("find_unique_constraint", trn.flavor, pg_schema=pg_schema)
            constraints = trn.execute(stmt()).fetchall()
            for tbl, constraint_name in constraints:
                col_stmt = Statement(
                    "find_index_columns",
                    trn.flavor,
                    pg_schema=pg_schema,
                    name=constraint_name,
                )
                columns = [c for (c,) in trn.execute(col_stmt())]
                # Postgresql will wrap columns names with quotes for
                # reserved words
                columns = [c.strip('"') for c in columns]
                by_constraint[tbl].append(columns)

        # Keep the unique constraint with the lowest number of columns for
        # each table
        res = {}
        for table, constraints in by_constraint.items():
            candidates = sorted(constraints, key=lambda item: len(item))
            for candidate in candidates:
                if len(candidate) == 1 and candidate[0] == db_pk.get(table):
                    # unique key is the primary key, skip
                    continue
                res[table] = candidate
                break
        return res

    def _create_views(self, trn: Transaction):
        # Create tables
        for name, view in self.views.items():
            if trn.flavor == "sqlite":
                # SQLite does not support OR REPLACE
                stmt = Statement(
                    "drop_view",
                    trn.flavor,
                    name=name,
                )
                yield stmt()

            stmt = Statement(
                "create_view",
                trn.flavor,
                name=name,
                view_def=view.view_def(),
            )
            yield stmt()

    def _create_tables(
        self, db_columns: dict[str, dict[str, str]], trn: Transaction | DummyTransaction
    ):
        # Create tables with all columns that are not foreign keys,
        # except for tables for which the primary key is also a foreign key.
        # These special tables need to be created after all other tables.
        tables = sorted(
            self.tables.values(), key=lambda t: t.primary_key in t.foreign_keys
        )
        for table in tables:
            if table.is_view:
                continue

            # table already exists
            if table.name in db_columns:
                continue

            ctypes = table.ctypes(trn.flavor, table.columns)
            if trn.flavor == "mssql" and table.has_array:
                warn(MSSQL_ARRAY_MSG.format(table=table.name), RuntimeWarning)
                continue

            columns_not_pk_fk = [
                col
                for col in table.columns
                if col != table.primary_key and col not in table.foreign_keys
            ]

            pk_fk_table = (
                self.tables.get(table.foreign_keys[table.primary_key])
                if table.primary_key in table.foreign_keys
                and table.primary_key is not None
                else None
            )

            # TODO use "KEY GENERATED ALWAYS AS IDENTITY" instead of
            # serials (see https://stackoverflow.com/a/55300741) ?
            stmt = Statement(
                "create_table",
                trn.flavor,
                table=table,
                columns=columns_not_pk_fk,
                ctypes=ctypes,
                not_null=table.not_null,
                default=table.default,
                pk_fk_table=pk_fk_table,
                natural_key=table.natural_key,
            )
            yield stmt()

    def _add_columns(
        self,
        db_columns: defaultdict[str, dict[str, str]],
        db_fks: defaultdict[str, dict[str, "FKConstraint"]],
        trn: Transaction | DummyTransaction,
    ):
        # Add remaining columns that were not created with the table
        # This concerns:
        # - columns that are foreign keys for tables that were just created
        # - columns added to existing tables
        for name, table in self.tables.items():
            if table.is_view:
                continue

            ctypes = table.ctypes(trn.flavor, table.columns)

            for col_name, col_type in ctypes.items():
                # if table was created by us, skip unless it's a foreign key that's not a primary key
                if name not in db_columns and (
                    col_name not in table.foreign_keys or col_name == table.primary_key
                ):
                    continue
                elif col_name in db_columns.get(name, {}):
                    # column already exists
                    # it is not a foreign key in the DB, but is declared as such in the schema
                    if (
                        col_name
                        not in [fk.column for fk in db_fks.get(name, {}).values()]
                        and col_name in table.foreign_keys
                    ):
                        raise RuntimeError(
                            f"Cannot add foreign key column '{col_name}' in table '{name}'."
                            " Foreign keys can only be added to newly created tables."
                        )
                    # column already exists, we skip
                    else:
                        continue

                fk_table = (
                    self.tables.get(table.foreign_keys[col_name])
                    if col_name in table.foreign_keys
                    else None
                )

                stmt = Statement(
                    "add_column",
                    trn.flavor,
                    table=name,
                    column=col_name,
                    col_def=col_type,
                    not_null=table.required(col_name),
                    default=table.default.get(col_name),
                    fk_table=fk_table,
                    primary_key=col_name == table.primary_key,
                )
                yield stmt()

    def _create_indexes(self, db_indexes, trn):
        # Add index on natural keys
        for name, table in self.tables.items():
            if table.is_view or f"{name}_idx" in db_indexes:
                continue
            if trn.flavor == "mssql" and table.has_array:
                continue
            if not table.natural_key:
                continue

            stmt = Statement(
                "create_unique_index",
                trn.flavor,
                table=name,
                natural_key=table.natural_key,
            )
            yield stmt()

    def setup_statements(self, trn: Optional[Transaction] = None):
        trn = trn or Transaction.current()
        # Find existing tables and columns
        db_columns = self._db_columns(trn)
        db_fks = self._db_fk(trn=trn)
        db_indexes = self._db_indexes(trn)

        yield from self._create_tables(db_columns, trn)
        yield from self._add_columns(db_columns, db_fks=db_fks, trn=trn)
        yield from self._create_indexes(db_indexes, trn)
        yield from self._create_views(trn)

    def create_tables(self, trn=None):
        """
        Create tables, indexes and foreign keys
        """
        trn = trn or Transaction.current()
        # Loop on setup statements and execute them
        for stm in self.setup_statements(trn=trn):
            trn.execute(stm)

    @classmethod
    def from_db(cls, trn: Optional[Transaction] = None) -> "Schema":
        """
        Instanciate a nagra Schema (and Tables) based on database
        schema
        """
        trn = trn or Transaction.current()
        schema = Schema()
        schema.introspect_db(trn=trn)
        return schema

    def introspect_db(self, *tables: str, trn: Optional[Transaction] = None):
        """
        Instanciate Table instances based on database content. If
        `tables` is non-empty, it is used as a whitelist and all other
        tables are ignored
        """
        from nagra.table import Table
        from nagra.view import View

        trn = trn or Transaction.current()
        db_fk = self._db_fk(*tables, trn=trn)
        db_pk = self._db_pk(trn=trn)
        db_unique = self._db_unique(db_pk, trn=trn)
        db_columns = self._db_columns(trn=trn)
        db_views = self._db_views(trn=trn)

        for table_name, cols in db_columns.items():
            if tables and table_name not in tables:
                continue
            fks = {fk.column: fk.foreign_table for fk in db_fk[table_name].values()}
            if view_def := db_views.get(table_name):
                # Instanciate view
                View(
                    table_name,
                    columns=cols,
                    natural_key=db_unique.get(table_name),
                    foreign_keys=fks,
                    schema=self,
                    as_select=view_def,
                )
            else:
                # Instanciate table
                Table(
                    table_name,
                    columns=cols,
                    natural_key=db_unique.get(table_name),
                    foreign_keys=fks,
                    primary_key=db_pk.get(table_name),
                    schema=self,
                )

    def drop(self, trn=None):
        trn = trn or Transaction.current()
        for table in self.tables.values():
            table.drop(trn)

    def generate_d2(self):
        tpl = template("misc/schema.d2")
        tables = self.tables.values()
        res = "\n".join(tpl.render(table=t) for t in tables)
        return res

    def generate_pydantic_models(
        self, base_class: str = "BaseModel", table_names: list[str] | None = None
    ):
        tpl = template("misc/pydantic-schema.py")
        if not table_names:
            tables = self.tables.values()
        else:
            tables = [self.tables[t] for t in table_names]

        res = "\n".join(
            tpl.render(
                table=t,
                class_name=snake_to_pascal(t.name),
                base_class=base_class,
                snake_to_pascal=snake_to_pascal,
            )
            for t in tables
        )
        return res

    def generate_toml(self):
        tpl = template("misc/schema-table.toml")
        tables = self.tables.values()

        res = "\n".join(tpl.render(
            table=t,
            skip_col=lambda c: c == "id" and t.primary_key == "id",
        ) for t in tables if not t.is_view)

        tpl = template("misc/schema-view.toml")
        views = self.views.values()
        res += "\n".join(tpl.render(view=v) for v in views)

        return res

    @contextmanager
    def suspend_fk(self, trn: Optional[Transaction] = None):
        """
        Temporarily drop all foreign keys and re-add them when
        exiting.  The db is introspected each time `suspend_fk` is
        called and the content of Schema is ignored, so the code may drop
        and re-add more foreign keys.
        """
        trn = trn or Transaction.current()
        if trn.flavor == "sqlite":
            trn.execute("PRAGMA foreign_keys = 0")
            yield
            trn.execute("PRAGMA foreign_keys = 1")
            return

        all_fks = list(
            chain.from_iterable(fks.values() for fks in self._db_fk(trn=trn).values())
        )
        for fk in all_fks:
            fk.drop()
        yield

        for fk in all_fks:
            fk.add()

    def eq(self, other):
        if not isinstance(other, Schema):
            return False
        checks = (
            sorted(self.tables) == sorted(other.tables),
            sorted(self.views) == sorted(other.views),
            # self.tables also contains views
            *(self.get(name).eq(other.get(name)) for name in self.tables),
        )
        return all(checks)


# Define default schema
Schema.default = Schema()


class FKConstraint:
    def __init__(self, name, table, column, foreign_table, foreign_column):
        self.name = name
        self.table = table
        self.column = column
        self.foreign_table = foreign_table
        self.foreign_column = foreign_column

    def drop(self, trn=None):
        trn = trn or Transaction.current()
        stmt = Statement(
            "drop_fk",
            trn.flavor,
            table=self.table,
            name=self.name,
        )
        trn.execute(stmt())

    def add(self, trn=None):
        trn = trn or Transaction.current()
        stmt = Statement(
            "add_foreign_key",
            trn.flavor,
            table=self.table,
            column=self.column,
            name=self.name,
            foreign_table=self.foreign_table,
            foreign_column=self.foreign_column,
        )
        trn.execute(stmt())



class BaseException(Exception):
    pass


class NoActiveTransaction(BaseException):
    def __init__(self, msg="No active transaction"):
        super().__init__(msg)


class TransactionReenterError(BaseException):
    pass


class UnresolvedFK(BaseException):
    pass


class ValidationError(BaseException):
    pass


class IncorrectSchema(BaseException):
    pass


class EvalTypeError(BaseException):
    pass



from nagra import View
from nagra.schema import Schema


def test_select_views(transaction, country, population, max_pop, min_pop):
    country.upsert("name").executemany(
        [
            ("Belgium",),
            ("Netherlands",),
        ]
    )

    population.upsert("country.name", "year", "value").executemany(
        [
            ("Belgium", 1970, 10),
            ("Belgium", 1971, 11),
            ("Netherlands", 1970, 12),
            ("Netherlands", 1971, 14),
        ]
    )
    # max_pop was defined by a subselect
    res = list(max_pop.select())
    assert res == [("Belgium", 11), ("Netherlands", 14)]

    res = list(
        max_pop.select(
            "max",
        ).orderby(
            ("country", "desc"),
        )
    )
    assert res == [(14,), (11,)]

    # min_pop was defined by nagra expressions
    res = list(min_pop.select())
    assert res == [("Belgium", 10), ("Netherlands", 12)]

    res = list(
        min_pop.select(
            "min",
        ).orderby(
            "country",
        )
    )
    assert res == [(10,), (12,)]


def test_get_view(min_pop, schema: Schema):
    assert min_pop == View.get("min_pop", schema=schema)



from nagra import Statement
from nagra.utils import strip_lines


def test_debug_statement():
    stmt = Statement("debug")
    doc = stmt()
    assert doc == ""

    stmt = stmt.foo("bar")
    doc = stmt()
    assert doc == "foo=bar\n"

    stmt = stmt.ham("spam")
    doc = stmt()
    assert doc == "foo=bar\nham=spam\n"


def test_mssql_select_limit():
    stmt = Statement(
        "select",
        flavor="mssql",
        table="people",
        columns=["[people].[name]"],
        joins=[],
        conditions=[],
        limit=5,
        offset=None,
        groupby=[],
        orderby=[],
        distinct_on=[],
        distinct=False,
    )
    lines = strip_lines(stmt())
    assert lines[0] == "SELECT TOP 5"
    assert lines[1] == "[people].[name]"
    assert lines[2] == "FROM [people]"


def test_mssql_select_offset_fetch():
    stmt = Statement(
        "select",
        flavor="mssql",
        table="people",
        columns=["[people].[name]"],
        joins=[],
        conditions=[],
        limit=10,
        offset=20,
        groupby=[],
        orderby=["[people].[name] ASC"],
        distinct_on=[],
        distinct=False,
    )
    doc = stmt()
    assert "ORDER BY" in doc
    assert "OFFSET 20 ROWS" in doc
    assert "FETCH NEXT 10 ROWS ONLY" in doc


def test_mssql_update_outputs_returning():
    stmt = Statement(
        "update",
        flavor="mssql",
        table="people",
        columns={"name": None, "id": None},
        condition_key=["id"],
        returning=["id"],
    )
    doc = stmt()
    assert "OUTPUT inserted.[id]" in doc


def test_mssql_upsert_uses_merge():
    stmt = Statement(
        "upsert",
        flavor="mssql",
        table="people",
        columns={"name": None, "email": None},
        conflict_key=["email"],
        do_update=False,
        returning=["id"],
        set_identity=False,
    )
    doc = stmt()
    lines = strip_lines(doc)
    assert lines == [
        'MERGE INTO [people] AS target',
        'USING',
        '(SELECT',
        '? AS [name], ? AS [email]',
        ') AS source',
        'ON (',
        'target.[email] = source.[email]',
        ')',
        'WHEN NOT MATCHED THEN INSERT ([name], [email]) VALUES (source.[name], '
        'source.[email])',
        'OUTPUT inserted.[id]',
        ';'
]




from typing import Optional
from datetime import datetime, date

from pydantic import BaseModel


def test_base_select(person):
    select = person.select("id", "name")
    model = select.to_pydantic(model_name="Person")

    class Person(BaseModel):
        id: int
        name: str

    assert model.model_json_schema() == Person.model_json_schema()


# def test_base_select_array(parameter):
#     select = parameter.select("name", "timestamps", "values")
#     model = select.to_pydantic(model_name="Parameter")
#     class Parameter(BaseModel):
#         name: str
#         timestamps: list[datetime]
#         values: list[float]
#     assert model.model_json_schema() == Parameter.model_json_schema()


def test_select_with_fk(person):
    select = person.select("id", "parent.name")
    model = select.to_pydantic(model_name="Person")

    class Person(BaseModel):
        id: int
        parent_name: Optional[str] = None

    assert  model.model_json_schema() == Person.model_json_schema()

    # Double fk
    select = person.select("id", "parent.parent.name")
    model = select.to_pydantic(model_name="Person")

    class Person(BaseModel):
        id: int
        parent_parent_name: Optional[str] = None

    assert model.model_json_schema() == Person.model_json_schema()


def test_kitchensink(kitchensink):
    select = kitchensink.select()
    aliases = kitchensink.columns
    model = select.to_pydantic(*aliases, model_name="KitchenSink")

    class KitchenSink(BaseModel):
        varchar: str
        bigint: Optional[int] = None
        float: Optional[float]
        int: Optional[int]
        timestamp: Optional[datetime] = None
        timestamptz: Optional[datetime] = None
        bool: Optional[bool] = None
        date: Optional[date]
        json: Optional[dict | list]  = None
        uuid: Optional[str] = None
        max: Optional[str] = None
        true: Optional[str] = None
        blob: Optional[bytes]

    # TODO loop on KitchenSink.model_fields and validate annotation and types




class {{class_name}}Stub({{base_class}}):
   {% for name in table.natural_key -%}
   {%- set col = table.columns[name] -%}
   {%- if name in table.foreign_keys -%}
         {{name}}: {{snake_to_pascal(name)}}Stub {% if not table.required(name) %} | None {% endif %}
   {%- else -%}
         {{name}}: {{col.dtype}} {% if not table.required(name) -%} | None {% endif %}
   {%- endif %}
   {% endfor %}

class {{class_name}}({{base_class}}):
    {% for name, col in table.columns.items() -%}
    {% if name in table.foreign_keys -%}
         {{name}}: {{snake_to_pascal(name)}}Stub {% if not table.required(name) -%} | None {% endif %}
    {% else -%}
         {{name}}: {{col.dtype}} {% if not table.required(name) -%} | None {% endif %}
    {% endif -%}
    {%- endfor -%}



from collections.abc import Iterable
from dataclasses import dataclass
from typing import Union, Optional, TYPE_CHECKING


from nagra import Statement, Schema
from nagra.exceptions import ValidationError
from nagra.transaction import Transaction
from nagra.upsert import Upsert
from nagra.writer import WriterMixin
from nagra.utils import snake_to_pascal, get_table_from_dataclass, iter_dataclass_cols


if TYPE_CHECKING:
    from nagra.table import Table


class Update(WriterMixin):
    def __init__(
        self,
        table: "Table",
        *columns: str,
        trn: Transaction,
        lenient: Union[bool, list[str], None] = None,
        check: Iterable[str] = [],
    ):
        self.table = table
        self.columns = list(columns)
        self.lenient = lenient or []
        self._check = list(check)
        self.trn = trn
        super().__init__()

    def clone(
        self,
        trn: Optional["Transaction"] = None,
        check: Iterable[str] = [],
    ):
        """
        Return a copy of update with updated parameters
        """
        trn = trn or self.trn
        check = self._check + list(check)
        cln = Update(
            self.table,
            *self.columns,
            trn=trn,
            lenient=self.lenient,
            check=check,
        )
        return cln

    def check(self, *conditions: str):
        return self.clone(check=conditions)

    def stm(self):
        pk = self.table.primary_key
        condition_key = [pk] if pk in self.groups else self.table.natural_key
        if not all(c in self.groups for c in condition_key):
            msg = "No unique key identifiable in provided columns !"
            # TODO put columns in msg
            raise ValidationError(msg)

        columns = self.groups
        stm = Statement(
            "update",
            self.trn.flavor,
            table=self.table.name,
            columns=columns,
            condition_key=condition_key,
            returning=[pk] if pk else self.table.natural_key,
        )
        return stm()

    def _exec_args(self, arg_df):
        # We need to reshuffle args values because they must be split
        # into SET and WHERE blocks in the sql statement
        pk = self.table.primary_key
        condition_key = [pk] if pk in self.groups else self.table.natural_key
        set_cols = [c for c in self.groups if c not in condition_key]
        sorted_groups = set_cols + condition_key
        args = zip(*(arg_df[c] for c in sorted_groups))
        return args

    @staticmethod
    def from_dataclass(cls: dataclass, schema: Schema = Schema.default) -> "Upsert":
        # TODO add from_pydantic
        table = get_table_from_dataclass(cls, schema)
        cols = list(iter_dataclass_cols(cls))
        return table.upsert(*cols)



from dataclasses import dataclass
from datetime import datetime, date
from typing import Optional, get_args

from nagra.select import clean_col, Select
from nagra.upsert import Upsert
from nagra.update import Update
from nagra.schema import Schema


def equivalent_classes(A, B):
    # Quickwin !
    if A == B:
        return True

    A_fields = vars(A)["__annotations__"]
    B_fields = vars(B)["__annotations__"]

    for A_name, B_name in zip(A_fields, B_fields):
        # Compare field names
        if A_name != B_name:
            breakpoint()
            return False

        # Compare field types
        A_type = A_fields[A_name]
        B_type = B_fields[B_name]
        if "__annotations__" in dir(A_type):
            return equivalent_classes(A_type, B_type)
        # We recurse on union types
        if get_args(A_type):
            for a_subtype, b_subtype in zip(get_args(A_type), get_args(B_type)):
                if not equivalent_classes(a_subtype, b_subtype):
                    return False
        elif A_type != B_type:
            breakpoint()
            return False

    return True


def test_base_select(person):
    select = person.select("id", "name")
    dclass = select.to_dataclass()

    @dataclass
    class Person:
        id: int
        name: str

    assert equivalent_classes(dclass, Person)


def test_base_select_array(parameter):
    select = parameter.select("name", "timestamps", "values")
    dclass = select.to_dataclass()

    @dataclass
    class Parameter:
        name: str
        timestamps: list[datetime] | None
        values: list[float] | None

    assert equivalent_classes(dclass, Parameter)


def test_select_with_fk(person):
    select = person.select("id", "parent.name")
    dclass = select.to_dataclass()

    @dataclass
    class Person:
        id: int
        parent_name: Optional[str]

    assert equivalent_classes(dclass, Person)

    # Double fk
    select = person.select("id", "parent.parent.name")
    dclass = select.to_dataclass()

    @dataclass
    class Person:
        id: int
        parent_parent_name: Optional[str]

    assert equivalent_classes(dclass, Person)


def test_select_with_sexp(person):
    select = person.select(
        "name",
        "(= name 'spam')",
        "(+ 1.0 1.0)",
        "(+ 2 2)",
    )
    dclass = select.to_dataclass("str_like", "bool_like", "float_like", "int_like")

    @dataclass
    class Expected:
        str_like: str
        bool_like: bool
        float_like: float
        int_like: int

    assert equivalent_classes(dclass, Expected)


def test_kitchensink(kitchensink):
    select = kitchensink.select()
    aliases = [c.upper() for c in kitchensink.columns]
    dclass = select.to_dataclass(*aliases)

    @dataclass
    class KitchenSink:
        VARCHAR: str
        BIGINT: Optional[int]
        FLOAT: Optional[float]
        INT: int
        TIMESTAMP: Optional[datetime]
        TIMESTAMPTZ: Optional[datetime]
        BOOL: Optional[bool]
        DATE: Optional[date]
        JSON: Optional[list | dict]
        UUID: Optional[str]
        MAX: Optional[str]
        TRUE: Optional[str]
        BLOB: Optional[bytes]

    assert equivalent_classes(dclass, KitchenSink)

    dclass = select.aliases(*aliases).to_dataclass()
    assert equivalent_classes(dclass, KitchenSink)


def test_aggregates(kitchensink):
    select = kitchensink.select(
        "(min varchar)",
        "(sum bigint)",
        "(avg float)",
        "(max int)",
        "(max timestamp)",
        "(date_bin '5 days' timestamptz)",
        "(now)",
        "(current_date)",
        "(date_part timestamp)",
        "(isfinite timestamp)",
        "(count)",
        "(every bool)",
        "(max date)",
    )
    dclass = select.to_dataclass(
        "varchar",
        "bigint",
        "float",
        "int",
        "timestamp",
        "timestamptz",
        "now",
        "current_date",
        "date_part",
        "isfinite",
        "count",
        "bool",
        "date",
    )

    @dataclass
    class KitchenSink:
        varchar: Optional[str]  # Aggregates can be null
        bigint: Optional[int]
        float: Optional[float]
        int: Optional[int]
        timestamp: Optional[datetime]
        timestamptz: Optional[datetime]
        now: datetime
        current_date: date
        date_part: Optional[float]
        isfinite: Optional[bool]
        count: Optional[int]
        bool: Optional[bool]
        date: Optional[date]

    assert equivalent_classes(dclass, KitchenSink)


def test_clean_col():
    assert clean_col("name") == "name"
    assert clean_col("table.name") == "table_name"
    assert clean_col("(= col 1)") == "___col_1_"


def test_nested_dataclasses(person):
    select = person.select("id", "parent.name")
    dclass = select.to_dataclass(nest=True)

    @dataclass
    class Parent:
        name: str

    @dataclass
    class Person:
        id: int
        parent: Parent | None

    assert equivalent_classes(dclass, Person)

    # Multiple fields for a fk
    select = person.select("id", "parent.name", "parent.id")
    dclass = select.to_dataclass(nest=True)

    @dataclass
    class Parent:
        name: str
        id: int

    @dataclass
    class Person:
        id: int
        parent: Parent | None

    assert equivalent_classes(dclass, Person)


def test_from_dataclass(person, schema: Schema):
    @dataclass
    class PersonLikeModel:
        __table__ = "person"
        name: str
        parent: str | None

    select = Select.from_dataclass(PersonLikeModel, schema=schema)
    assert list(select.columns) == ["name", "parent"]

    @dataclass
    class PersonStub:
        __table__ = "person"
        name: str

    @dataclass
    class Person:
        name: str
        parent: PersonStub

    select = Select.from_dataclass(Person, schema=schema)
    assert list(select.columns) == ["name", "parent.name"]
    assert select.table.name == "person"

    for cls in [Upsert, Update]:
        obj = cls.from_dataclass(Person, schema=schema)
        assert list(obj.columns) == ["name", "parent.name"]
        assert obj.table.name == "person"



import dataclasses
from collections import defaultdict
from collections.abc import Iterable
from functools import partial
from itertools import islice
from typing import TYPE_CHECKING

from nagra.exceptions import UnresolvedFK, ValidationError
from nagra.utils import logger
from nagra.transaction import ExecMany


if TYPE_CHECKING:
    from pandas import DataFrame
    from polars import LazyFrame


class WriterMixin:
    """
    Utility class that provide common methods for Update and Upsert
    """

    def __init__(self):
        self.groups, self.resolve_stm = self.prepare()

    def prepare(self):
        """
        Organise columns in groups and prepare statement to
        resolve fk based on columns expressions
        """
        groups = defaultdict(list)
        for col in self.columns:
            if "." in col:
                head, tail = col.split(".", 1)
                groups[head].append(tail)
            else:
                groups[col] = None

        resolve_stm = {}
        for col, to_select in groups.items():
            if not to_select:
                continue
            cond = ["(= %s {})" % c for c in to_select]
            ftable = self.table.schema.get(self.table.foreign_keys[col])
            select = ftable.select(ftable.primary_key).where(*cond)
            resolve_stm[col] = select.stm()
        return groups, resolve_stm

    def execute(self, *values):
        ids = self.executemany([values])
        if ids:
            return ids[0]

    def executemany(self, records: Iterable[tuple]) -> list:
        # Transform list of records into a dataframe-like dict
        value_df = dict(zip(self.columns, zip(*records)))
        if not value_df:
            return []
        arg_df = {}
        for col, to_select in self.groups.items():
            if to_select:
                values = list(zip(*(value_df[f"{col}.{s}"] for s in to_select)))
                # Try to instanciate lru cache
                cache_key = (self.resolve_stm[col], str(self.lenient))
                lru = self.trn.get_fk_cache(cache_key, fn=partial(self._resolve, col))
                if lru is not None:
                    arg_df[col] = lru.run(values)
                else:
                    arg_df[col] = self._resolve(col, values)
            else:
                arg_df[col] = value_df[col]

        # Build arg iterable
        args = self._exec_args(arg_df)
        # Work by chunks
        stm = self.stm()
        ids = []
        returning = self.table.primary_key is not None
        while True:
            chunk = list(islice(args, 1000))
            if not chunk:
                break
            match self.trn.flavor:
                case "sqlite" | "mssql":
                    for item in chunk:
                        cursor = self.trn.execute(stm, item)
                        if returning:
                            new_id = cursor.fetchone()
                            ids.append(new_id[0] if new_id else None)
                        cursor.close()

                case "postgresql":
                    cursor = self.trn.executemany(stm, chunk, returning)
                    if returning:
                        ids.extend(r and r[0] for r in cursor)

        # If conditions are present, enforce those
        if self._check:
            self.validate(ids)
        return ids

    def validate(self, ids: list[int]):
        iter_ids = iter(ids)
        pk = self.table.primary_key
        while True:
            chunk = list(islice(iter_ids, 1000))
            if not chunk:
                return
            cond = self._check + [f"(in {pk} %s)" % (" {}" * len(chunk))]
            select = self.table.select("(count *)").where(*cond)
            (count,) = select.execute(*chunk).fetchone()
            if count != len(chunk):
                msg = f"Validation failed! Condition is: {self._check} )"
                raise ValidationError(msg)

    def _resolve(self, col, values):
        # XXX Detect situation where more than on result is found for
        # a given value (we could also enforce that we only resolve
        # columns with unique constraints) ?
        stm = self.resolve_stm[col]
        exm = ExecMany(stm, values, trn=self.trn)
        for res, vals in zip(exm, values):
            if res is not None:
                yield res[0]
            elif any(v is None for v in vals):
                # One of the values is not given
                yield None
            elif self.lenient is True or col in self.lenient:
                msg = "Value '%s' not found for foreign key column '%s' of table %s"
                logger.info(msg, vals, col, self.table)
                yield None
            else:
                raise UnresolvedFK(
                    f"Unable to resolve '{vals}' (for foreign key "
                    f"{col} of table {self.table.name})"
                )

    def __call__(self, records):
        return self.executemany(records)

    def from_pandas(self, df: "DataFrame"):
        if df.empty:
            return self.executemany([])

        # Convert non-basic types to string
        is_copy = False
        for col in self.columns:
            if df[col].dtype in ("int", "float", "bool", "str"):
                continue
            if df[col].dtype == "object":
                # bytes is not a dedicated type, we rely on the first
                # value and hope the column type is consistent
                if isinstance(df[col][0], bytes):
                    continue
            if not is_copy:
                df = df.copy()
                is_copy = True
            df[col] = df[col].astype(str)

        rows = df[self.columns].values
        return self.executemany(rows)

    def from_polars(self, df: "LazyFrame", batch: bool = False):
        """
        Write data from a polars LazyFrame. Set `batch` to True
        to enable streaming through the collect_batches() method
        """
        from polars import Struct, col

        # Ignore extra columns
        df.select(self.columns)
        schema = df.collect_schema()
        # Convert non-basic types to string
        for name, dtype in schema.items():
            if dtype == Struct:
                df = df.with_columns(col(name).struct.json_encode())

        if batch:
            chunks = df.collect_batches()
        else:
            chunks = [df.collect()]

        res = []
        for chunk in chunks:
            rows = chunk.iter_rows()
            res += self.executemany(rows)
        return res

    def from_dict(self, records):
        # Create select object in order to generate the same column names
        select = self.table.select(*self.columns)
        field_names = [f.name for f in dataclasses.fields(select.to_dataclass())]

        # Extract dict values - allows for field names or dotted column format
        f_or_c = list(zip(field_names, self.columns))
        rows = (
            tuple(getter(record, field, col) for col, field in f_or_c)
            for record in records
        )
        return self.executemany(rows)


def getter(record, field, col):
    """
    Utility function to extract value as a column name (with
    potentially dots) or a field name from record
    """
    if field in record:
        return record[field]
    if col in record:
        return record[col]
    raise KeyError(f"KeyError: neither {field} or {col} found")


def _slicer(chunk_size=10_000):
    start = 0
    while True:
        stop = start + chunk_size
        yield start, stop
        start = stop



import zoneinfo
from datetime import datetime, date
from pandas import concat, DataFrame, to_datetime
from uuid import UUID

import pytest

from nagra import Transaction


def test_to_pandas(transaction, temperature):
    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    # Read data
    df = temperature.select().to_pandas()
    assert list(df.columns) == ["timestamp", "city", "value"]
    assert sorted(df.city) == ["Berlin", "London"]

    # Read data - with chunks
    dfs = temperature.select().to_pandas(chunked=1)
    df = concat(list(dfs))
    assert list(df.columns) == ["timestamp", "city", "value"]
    assert sorted(df.city) == ["Berlin", "London"]

    # Read with custom arg
    cond = "(= value {})"
    df = temperature.select().where(cond).to_pandas(12)
    assert list(df.columns) == ["timestamp", "city", "value"]
    assert sorted(df.city) == ["London"]


def test_from_pandas(transaction, kitchensink):
    if transaction.flavor == "mssql":
        pytest.skip("TZ-aware timestamps are not supported by MSSQL")

    df = DataFrame(
        {
            "varchar": ["ham"],
            "bigint": [1],
            "float": [1.0],
            "int": [1],
            "timestamp": to_datetime(["1970-01-01 00:00:00"]),
            "timestamptz": to_datetime(["1970-01-01 00:00:00+00:00"]),
            "bool": [True],
            "date": ["1970-01-01"],
            "json": [{}],
            "uuid": ["F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"],
            "max": ["max"],
            "true": ["true"],
            "blob": [b"blob"],
        }
    )

    # UPSERT
    kitchensink.upsert().from_pandas(df)
    (row,) = kitchensink.select()
    BRUTZ = zoneinfo.ZoneInfo(key="Europe/Brussels")
    if Transaction.current().flavor == "postgresql":
        assert row == (
            "ham",
            1,
            1.0,
            1,
            datetime(1970, 1, 1, 0, 0),
            datetime(1970, 1, 1, 1, 0, tzinfo=BRUTZ),
            True,
            date(1970, 1, 1),
            {},
            UUID("F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"),
            "max",
            "true",
            b"blob",
        )
    else:
        assert row == (
            "ham",
            1,
            1.0,
            1,
            "1970-01-01",
            "1970-01-01 00:00:00+00:00",
            1,
            "1970-01-01",
            "{}",
            "F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C",
            "max",
            "true",
            b"blob",
        )

    # SELECT with operator
    if transaction.flavor == "postgresql":
        new_df = kitchensink.select(
            "(date_bin '5 days' timestamptz '1900-01-01')",
        ).to_pandas()
        new_df.columns = ["ts"]
        assert str(new_df.ts.dtype) == "datetime64[ns, Europe/Brussels]"
        ts = new_df.ts[0]
        assert ts.isoformat() == "1969-12-30T01:00:00+01:00"
        # NOTE the above result is expected:
        # ```
        # =# SELECT date_bin('5 days', TIMESTAMPTZ '1970-01-01 00:00:00+00', '1900-01-01');
        #       date_bin
        # ------------------------
        #  1969-12-30 01:00:00+01
        # (1 row)
        # ```

    if transaction.flavor == "sqlite":
        new_df = kitchensink.select(
            "(+ float float)",
            "(- int int)",
        ).to_pandas()
        new_df.columns = ["float", "int"]
        assert str(new_df.float.dtype) == "float64"
        assert str(new_df.int.dtype) == "int64"



from nagra.sexpr import AST
from nagra.table import Table, Env
from nagra.utils import strip_lines


def test_sexpr():
    # Simple dot reference
    expr = "ham.spam"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<VarToken ham.spam>]"

    # With int literal
    expr = "(= ham.spam 1)"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<BuiltinToken =>, <VarToken ham.spam>, <IntToken 1>]"

    # With string literal
    expr = "(= ham.spam 'one')"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<BuiltinToken =>, <VarToken ham.spam>, <StrToken one>]"

    # With placeholder
    expr = "(= ham.spam {})"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<BuiltinToken =>, <VarToken ham.spam>, <ParamToken >]"

    # With string operator
    expr = "(|| 'one' 'two')"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<BuiltinToken ||>, <StrToken one>, <StrToken two>]"

    # With variable clashing with an aggregate
    expr = "(max min)"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<AggToken max>, <VarToken min>]"

    # With lone aggregate
    expr = "(count)"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<AggToken count>]"

    # With an operator sign as variable
    expr = "(+ 1 1)"
    ast = AST.parse(expr)
    assert str(ast.tokens) == "[<BuiltinToken +>, <IntToken 1>, <IntToken 1>]"

    # With litterals
    expr = "(is null true)"
    ast = AST.parse(expr)
    assert (
        str(ast.tokens)
        == "[<BuiltinToken is>, <LiteralToken null>, <LiteralToken true>]"
    )

    # Use dot prefix to escape literal
    expr = "(is null .true))"
    ast = AST.parse(expr)
    assert (
        str(ast.tokens) == "[<BuiltinToken is>, <LiteralToken null>, <VarToken true>]"
    )

    # Compare litterals
    expr = "(!= true false)"
    ast = AST.parse(expr)
    assert (
        str(ast.tokens)
        == "[<BuiltinToken !=>, <LiteralToken true>, <LiteralToken false>]"
    )


def test_find_relations():
    expr = "(= ham.spam foo.bar)"
    ast = AST.parse(expr)
    assert list(ast.relations()) == ["ham.spam", "foo.bar"]


def test_simple_eval():
    table = Table("spam", {"a": "bool"})
    expr = "a"
    ast = AST.parse(expr)
    env = Env(table)
    assert ast.eval(env) == '"spam"."a"'

    expr = "(= a 1)"
    ast = AST.parse(expr)
    env = Env(table)
    assert ast.eval(env) == '"spam"."a" = 1'

    expr = "(= a (= 1 1))"
    ast = AST.parse(expr)
    env = Env(table)
    assert ast.eval(env) == '"spam"."a" = (1 = 1)'

    # AND and OR, OR add extra parents in evaled form
    expr = "(and true true)"
    ast = AST.parse(expr)
    env = Env(table)
    assert ast.eval(env) == "true AND true"

    expr = "(or true true)"
    ast = AST.parse(expr)
    env = Env(table)
    assert ast.eval(env) == "(true OR true)"

    expr = "(or (and true false) true)"
    ast = AST.parse(expr)
    env = Env(table)
    assert ast.eval(env) == "((true AND false) OR true)"


def test_join_eval(person):
    # Unique  join
    env = Env(table=person)
    expr = "(and (= parent.name 'Roger')"
    ast = AST.parse(expr)
    res = ast.eval(env)
    assert env.refs == {
        ("parent",): "parent_0",
    }
    assert res == """("parent_0"."name" = 'Roger')"""

    # Double join with two different depth
    env = Env(table=person)
    expr = "(and (= parent.name 'Roger') (= parent.parent.name 'George')"
    ast = AST.parse(expr)
    res = ast.eval(env)
    assert env.refs == {
        ("parent",): "parent_0",
        ("parent", "parent"): "parent_1",
    }
    assert res == """("parent_0"."name" = 'Roger') AND ("parent_1"."name" = 'George')"""

    # Double join with same depth
    env = Env(table=person)
    expr = "(and (= parent.name 'Roger') (= parent.id 1)"
    ast = AST.parse(expr)
    res = ast.eval(env)
    assert env.refs == {
        ("parent",): "parent_0",
    }
    assert res == """("parent_0"."name" = 'Roger') AND ("parent_0"."id" = 1)"""


def test_eval_dtype(kitchensink):
    env = Env(table=kitchensink)
    expr = "(is blob null)"
    ast = AST.parse(expr)
    res = ast.eval_type(env)
    assert res is bool


def test_dynamic_builtins(kitchensink):
    """
    If an opetator is not known, it is considerer as a prefix
    operatore (function call-lile syntax).
    """
    select = kitchensink.select(
        "(date_bin '5 days' timestamp '2025-01-01')", "(avg value)"
    )
    res = strip_lines(select.stm())
    assert res == [
        "SELECT",
        "date_bin('5 days', \"kitchensink\".\"timestamp\", '2025-01-01'), "
        'avg("kitchensink"."value")',
        'FROM "kitchensink"',
        "GROUP BY",
        "date_bin('5 days', \"kitchensink\".\"timestamp\", '2025-01-01')",
        ";",
    ]


def test_is_nullable(person):
    env = Env(table=person)

    # Aggregates are nullable (on empty tables)
    expr = "(max id)"
    ast = AST.parse(expr)
    assert ast.is_nullable(env)

    # required column are not
    expr = "name"
    ast = AST.parse(expr)
    assert not ast.is_nullable(env)


def test_single_token(person):
    """
    Check AST behavour with single token expressions (no
    parenthesis)
    """
    env = Env(table=person)

    expr = "'Alice'"
    ast = AST.parse(expr)
    res = ast.eval(env)
    assert res == "'Alice'"

    expr = "0.0"
    ast = AST.parse(expr)
    res = ast.eval(env)
    assert res == 0
    assert ast.eval_type(env) is float

    expr = "parent"
    ast = AST.parse(expr)
    res = ast.eval(env)
    assert res == '"person"."parent"'
    assert ast.eval_type(env) is int



from nagra.utils import strip_lines


def test_delete(person):
    # Simple delete
    delete = person.delete()
    stm = delete.stm()
    res = strip_lines(stm)
    assert res == ['DELETE FROM "person"']

    # With a condition
    delete = person.delete('(= name "spam")')
    stm = delete.stm()
    res = strip_lines(stm)
    assert res == ['DELETE FROM "person"', "WHERE", '"person"."name" = \'spam\'']

    # With a join
    delete = person.delete('(= parent.name "spam")')
    stm = delete.stm()
    res = strip_lines(stm)
    assert res == [
        'DELETE FROM "person"',
        'WHERE "person".id IN (',
        'SELECT "person".id from "person"',
        'LEFT JOIN "person" as parent_0 ON (',
        'parent_0."id" = "person"."parent"',
        ")WHERE",
        '"parent_0"."name" = \'spam\'',
        ")",
    ]


def test_delete_cascade(transaction, person, skill):
    # Insert persons
    person.upsert("name").executemany(
        [
            ("Yankee",),
            ("Zulu",),
        ]
    )

    # Add skills (person is a fk and is not null)
    skill.upsert("name", "person.name").executemany(
        [
            ("Cooking", "Yankee"),
            ("Fishing", "Zulu"),
        ]
    )
    # Check created skills
    assert sorted(skill) == [("Cooking", "Yankee"), ("Fishing", "Zulu")]

    # Delete person and list skills
    person.delete('(= name "Zulu")').execute()
    assert list(skill) == [("Cooking", "Yankee")]

    # Same, but with an  arg
    person.delete("(= name {})").execute("Yankee")
    assert list(skill) == []



from datetime import datetime

import pytest

from nagra.utils import strip_lines
from nagra.exceptions import UnresolvedFK, ValidationError


def test_simple_upsert_stm(person):
    upsert = person.upsert("name", "parent.name")
    assert list(upsert.resolve_stm) == ["parent"]
    res = strip_lines(upsert.resolve_stm["parent"])
    assert res == [
        "SELECT",
        '"person"."id"',
        'FROM "person"',
        "WHERE",
        '"person"."name" = %s',
        ";",
    ]
    res = strip_lines(upsert.stm())
    assert res == [
        'INSERT INTO "person" ("name", "parent")',
        "VALUES (",
        "%s,%s",
        ")",
        "ON CONFLICT (",
        '"name"',
        ")",
        "DO UPDATE SET",
        '"parent" = EXCLUDED."parent"',
        'RETURNING "id"',
    ]


def test_simple_upsert(cacheable_transaction, person):
    # First upsert
    upsert = person.upsert("name")
    upsert.execute("Big Bob")
    (record,) = list(person.select("name").execute())
    assert record == ("Big Bob",)

    # Second one
    upsert = person.upsert("name", "parent.name")
    upsert.execute("Bob", "Big Bob")
    rows = list(person.select("name", "parent.name").orderby("name").execute())
    assert rows == [("Big Bob", None), ("Bob", "Big Bob")]

    # Empty input
    upsert = person.upsert("name", "parent.name")
    upsert.execute()
    upsert.executemany([])
    rows = list(person.select("name", "parent.name").orderby("name").execute())
    assert rows == [("Big Bob", None), ("Bob", "Big Bob")]


def test_insert(cacheable_transaction, person):
    # First upsert
    upsert = person.upsert("name")
    records = [("Big Bob",), ("Bob",)]
    upsert.executemany(records)

    # Second one (with insert instead of upsert)
    upsert = person.insert("name", "parent.name")
    upsert.execute("Bob", "Big Bob")
    rows = list(person.select("name", "parent.name").orderby("name").execute())
    assert rows == [("Big Bob", None), ("Bob", None)]


def test_insert_no_nk_pk(cacheable_transaction, temperature_no_nk_pk):
    # First simple insert - should work even without nk/pk
    insert = temperature_no_nk_pk.insert("timestamp", "city", "value")
    records = [
        (datetime(1970, 1, 1, 0, 0), "Berlin", 10.0),
        (datetime(1970, 1, 2, 0, 0), "London", 12.0),
    ]
    insert.executemany(records)
    if cacheable_transaction.flavor == "sqlite":
        records = [
            tuple(str(v) if isinstance(v, datetime) else v for v in r) for r in records
        ]
    rows = list(temperature_no_nk_pk.select().orderby("city").execute())
    assert rows == records

    # Upsert should fail without nk/pk
    upsert = temperature_no_nk_pk.upsert("timestamp", "city", "value")
    with pytest.raises(ValidationError):
        upsert.execute(datetime(1970, 1, 2, 0, 0), "Berlin", 11.0)


def test_upsert_stmt_with_id(cacheable_transaction, person):
    if cacheable_transaction.flavor == "postgresql":
        # Test stmt with all columns
        upsert = person.upsert("id", "name", "parent.name")
        res = list(strip_lines(upsert.stm()))
        assert res == [
            'INSERT INTO "person" ("id", "name", "parent")',
            "VALUES (",
            "%s,%s,%s",
            ")",
            "ON CONFLICT (",
            '"id"',
            ")",
            "DO UPDATE SET",
            '"name" = EXCLUDED."name" , "parent" = EXCLUDED."parent"',
            'RETURNING "id"',
        ]

        # Test stmt with one columns
        upsert = person.upsert("id", "name")
        res = list(strip_lines(upsert.stm()))
        assert res == [
            'INSERT INTO "person" ("id", "name")',
            "VALUES (",
            "%s,%s",
            ")",
            "ON CONFLICT (",
            '"id"',
            ")",
            "DO UPDATE SET",
            '"name" = EXCLUDED."name"',
            'RETURNING "id"',
        ]

    # Insert & update on db
    upsert = person.upsert("name")
    new_id = upsert.execute("Lima")

    upsert = person.upsert("id", "name")
    upsert.execute(new_id, "Lima2")
    (record,) = person.select("id", "name").where("(= id {})").execute(new_id)
    assert record == (new_id, "Lima2")


def test_upsert_exec_with_id(cacheable_transaction, person):
    # Add parent
    upsert = person.upsert("id", "name")
    upsert.execute(1, "Big Bob")
    (rows,) = list(person.select("id", "name").execute())
    assert rows == (1, "Big Bob")

    # Add child
    upsert = person.upsert("id", "name", "parent.name")
    upsert.execute(2, "Bob", "Big Bob")
    rows = list(person.select("name", "parent.name").orderby("name").execute())
    assert rows == [("Big Bob", None), ("Bob", "Big Bob")]

    # Update child
    upsert = person.upsert("id", "name")
    upsert.execute(2, "BOB")
    cond = "(= id 2)"
    (rows,) = person.select("name").where(cond).execute()
    assert rows == ("BOB",)


def test_many_upsert(cacheable_transaction, person):
    # First upsert
    upsert = person.upsert("name")
    records = [("Big Alice",), ("Big Bob",)]
    upsert.executemany(records)
    rows = list(person.select("name").execute())
    assert len(rows) == 2

    # Second upsert
    upsert = person.upsert("name", "parent.name")
    records = [
        (
            "Alice",
            "Big Alice",
        ),
        (
            "Bob",
            "Big Bob",
        ),
    ]
    upsert.executemany(records)

    rows = list(person.select("name", "parent.name").execute())
    assert len(rows) == 4


def test_dbl_fk_upsert(cacheable_transaction, person):
    # GP
    upsert = person.upsert("name")
    records = [("GP Alice",), ("GP Bob",)]
    upsert.executemany(records)

    # Parents
    upsert = person.upsert("name", "parent.name")
    records = [
        (
            "P Alice",
            "GP Alice",
        ),
        (
            "P Bob",
            "GP Bob",
        ),
    ]
    upsert.executemany(records)

    # children
    upsert = person.upsert("name", "parent.parent.name")
    records = [
        (
            "Alice",
            "GP Alice",
        ),
        (
            "Bob",
            "GP Bob",
        ),
    ]
    upsert.executemany(records)

    select = (
        person.select(
            "name",
            "parent.name",
            "parent.parent.name",
        )
        .where("(not (is parent.parent.name null))")
        .orderby("name")
    )
    rows = list(select)
    assert rows == [
        ("Alice", "P Alice", "GP Alice"),
        ("Bob", "P Bob", "GP Bob"),
    ]


def test_missing_fk(cacheable_transaction, person):
    # If pass None in parent.name, we get None back
    upsert = person.upsert("name", "parent.name")
    records = [("Big Alice", None), ("Big Bob", None)]
    upsert.executemany(records)

    rows = list(person.select("parent").execute())
    assert rows == [(None,), (None,)]

    # If given a non-existing name upsert raises UnresolvedFK exception
    upsert = person.upsert("name", "parent.name")
    records = [("Big Alice", "I do not exist")]
    with pytest.raises(UnresolvedFK):
        upsert.executemany(records)

    # If lenient is given a None is inserted
    for lenient in [True, ["parent"]]:
        upsert = person.upsert("name", "parent.name", lenient=lenient)
        records = [("Big Alice", "I do not exist")]
        upsert.executemany(records)
        rows = list(person.select("parent").where("(= name 'Big Alice')").execute())
        assert rows == [(None,)]


def test_resolve(cacheable_transaction, person):
    # Prepare table
    upsert = person.upsert("name", "parent.name")
    records = [("Big Alice", None), ("Big Bob", None)]
    upsert.executemany(records)

    # Direct call to resolve, without upserting anything
    values = ["Big Alice", "Big Bob"]
    ids = list(upsert.resolve("parent", values))
    assert ids == [1, 2]


def test_return_ids(cacheable_transaction, person):
    # Create an "on conflict update" upsert
    upsert = person.upsert("name", "parent.name")
    records = [("Big Alice", None), ("Big Bob", None)]
    insert_ids = upsert.executemany(records)
    update_ids = upsert.executemany(records)
    assert len(insert_ids) == 2
    assert insert_ids == update_ids
    assert insert_ids != [None, None]

    # Create an "on conflict do nothing" upsert (because we only write
    # nk)
    upsert = person.upsert("name")
    records = [("Papa",), ("Quebec",)]
    insert_ids = upsert.executemany(records)
    assert insert_ids != [None, None]
    update_ids = upsert.executemany(records)
    assert update_ids == [None, None]

    # This time pass other columns but with an insert
    records = [("Papa", 1), ("Quebec", 1)]
    insert_ids = person.insert("name", "parent").executemany(records)
    assert insert_ids == [None, None]

    # Check db content
    res = list(person.select("parent").where("(= name {})").execute("Papa"))
    assert res == [(None,)]


def test_double_insert(cacheable_transaction, person):
    """
    Show that 'last write win' when duplicates are given
    """
    upsert = person.upsert("name")
    upsert.execute("Tango")

    upsert = person.upsert("name", "parent.name")
    upsert.executemany(
        [
            ("Charly", "Tango"),
            ("Charly", None),
        ]
    )
    rows = list(person.select().orderby(("name", "desc")))
    assert rows == [("Tango", None), ("Charly", None)]


def test_one2many_ref(cacheable_transaction, person, org):
    person.upsert("name").execute("Charly")
    person.upsert("name").execute("Juliet")
    org.upsert("name", "person.name").execute("Alpha", "Charly")
    org.upsert("name", "person.name").execute("Bravo", "Juliet")

    # update parent based on org
    upsert = person.upsert("name", "parent.orgs.name")
    upsert.execute("Juliet", "Alpha")

    # Check results
    rows = list(person.select().where("(= name 'Juliet')"))
    assert rows == [("Juliet", "Charly")]


def test_where_cond(cacheable_transaction, person):
    """
    Shows that an exception is raised when a row infrige a check condition
    """
    upsert = person.upsert("name")
    upsert.execute("Tango")

    cond = "(!= name parent.name)"  # Forbid self-reference
    upsert = person.upsert("name", "parent.name").check(cond)
    with pytest.raises(ValidationError):
        upsert.execute("Tango", "Tango")


def test_default_value(transaction, org):
    """
    Shows that default values are applied on row creation
    """
    upsert = org.upsert("name")
    upsert.execute("Lima")

    (record,) = org.select("name", "status")
    name, status = record
    assert (name, status) == ("Lima", "OK")


def test_mixed_cursor(cacheable_transaction, person):
    if cacheable_transaction.flavor == "mssql":
        pytest.skip("MSSQL does not support multiple live cursors")

    # First upsert
    upsert = person.upsert("name")
    records = [("Romeo",), ("Sierra",), ("Tango",)]
    upsert.executemany(records)

    # add Tango as parent to other record, execute one stm for each
    select = person.select("name").where("(!= name 'Tango')")
    upsert = person.upsert("name", "parent.name")
    for (name,) in select:
        # In this loop we have two "live" cursor, the one consumed by
        # the select and the one executed by the upsert
        upsert.execute(name, "Tango")

    # assert results
    select = person.select("name", "parent.name").where("(!= name 'Tango')")
    for name, parent in select:
        assert name in ["Romeo", "Sierra"]
        assert parent == "Tango"


def test_arrays(cacheable_transaction, parameter):
    if cacheable_transaction.flavor == "mssql":
        pytest.skip("MSSQL does not support array columns")
    # First upsert
    upsert = parameter.upsert()
    records = [
        ("one", ["2024-08-03T10:42", "2024-08-03T10:43"], [2, 3]),
        ("two", ["2024-08-03T10:44", "2024-08-03T10:45"], [4, 5]),
        ("three", ["2024-08-03T10:46", "2024-08-03T10:47"], [6, 7]),
    ]
    if cacheable_transaction.flavor == "sqlite":
        records = [[str(v) for v in r] for r in records]
    upsert.executemany(records)

    records = list(parameter.select().orderby("id"))

    if cacheable_transaction.flavor == "sqlite":
        assert records == [
            ("one", "['2024-08-03T10:42', '2024-08-03T10:43']", "[2, 3]"),
            ("two", "['2024-08-03T10:44', '2024-08-03T10:45']", "[4, 5]"),
            ("three", "['2024-08-03T10:46', '2024-08-03T10:47']", "[6, 7]"),
        ]
    else:
        assert records == [
            (
                "one",
                [datetime(2024, 8, 3, 10, 42), datetime(2024, 8, 3, 10, 43)],
                [2.0, 3.0],
            ),
            (
                "two",
                [datetime(2024, 8, 3, 10, 44), datetime(2024, 8, 3, 10, 45)],
                [4.0, 5.0],
            ),
            (
                "three",
                [datetime(2024, 8, 3, 10, 46), datetime(2024, 8, 3, 10, 47)],
                [6.0, 7.0],
            ),
        ]


def test_from_dict(transaction, person):
    # Prepare two lines in table
    upsert = person.upsert("name", "parent.name")
    upsert.execute("Big Bob", None)
    upsert.execute("Bob", "Big Bob")
    records = list(person.select().to_dict())
    assert records == [
        {"name": "Big Bob", "parent_name": None},
        {"name": "Bob", "parent_name": "Big Bob"},
    ]

    # Upserting unchanged data is a noop
    person.upsert("name", "parent.name").from_dict(records)
    records_bis = list(person.select().to_dict())
    assert records_bis == records

    # Upserting will also support dotted notation
    new_records = [
        {"name": "Big Bob", "parent.name": None},
        {"name": "Other Bob", "parent.name": "Big Bob"},
    ]
    person.upsert("name", "parent.name").from_dict(new_records)
    records_bis = list(person.select().orderby("id").to_dict())
    assert records_bis == [
        {"name": "Big Bob", "parent_name": None},
        {"name": "Bob", "parent_name": "Big Bob"},
        {"name": "Other Bob", "parent_name": "Big Bob"},
    ]


def test_nk_less_table(transaction, value):
    # Simple insert
    upsert = value.insert("value")
    ids = upsert.execute(1)
    assert ids == 1

    # Update work if id is given
    upsert = value.update("id", "value")
    ids = upsert.execute(1, 1.1)
    assert ids == 1

    # Upsert should fail if id is not given
    upsert = value.upsert("value")
    with pytest.raises(ValidationError):
        upsert.execute(1)
    # return



from nagra import Statement
from nagra.utils import strip_lines
from nagra.schema import Schema
from nagra.table import Table


def test_create_table(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    Table(
        "my_table",
        columns={
            "name": "varchar",
            "score": "int",
        },
        natural_key=["name"],
        not_null=["score"],
        default={
            "score": "0",
        },
        primary_key="custom_id",
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))
    create_table, create_idx = map(strip_lines, lines)

    match flavor:
        case "postgresql":
            assert create_table == [
                'CREATE TABLE  "my_table" (',
                '"custom_id" BIGSERIAL PRIMARY KEY',
                ",",
                '"name" TEXT',
                "NOT NULL",
                ",",
                '"score" INTEGER',
                "NOT NULL",
                "DEFAULT 0",
                ");",
            ]
        case "sqlite":
            assert create_table == [
                'CREATE TABLE  "my_table" (',
                '"custom_id" INTEGER PRIMARY KEY',
                ",",
                '"name" TEXT',
                "NOT NULL",
                ",",
                '"score" INTEGER',
                "NOT NULL",
                "DEFAULT 0",
                ");",
            ]
        case "mssql":
            assert create_table == [
                "CREATE TABLE [my_table] (",
                "[custom_id] BIGINT IDENTITY(1,1) PRIMARY KEY",
                ",",
                "[name] NVARCHAR(200)",
                "NOT NULL",
                ",",
                "[score] INT",
                "NOT NULL",
                "DEFAULT 0",
                ");",
            ]
    match flavor:
        case "postgresql" | "sqlite":
            assert create_idx == [
                'CREATE UNIQUE INDEX my_table_idx ON "my_table" (',
                '"name"',
                ");",
            ]
        case "mssql":
            assert create_idx == [
                "CREATE UNIQUE INDEX [my_table_idx] ON [my_table] ([name]",
                ");",
            ]


def test_create_table_pk_is_fk(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    # Table with no natural key and a fk in the primary key
    # It should be created second
    Table(
        "score",
        columns={
            "concept": "bigint",
            "score": "int",
        },
        primary_key="concept",
        foreign_keys={
            "concept": "concept",
        },
        schema=schema,
    )
    Table(  # Regular table
        "concept",
        columns={
            "name": "varchar",
        },
        natural_key=["name"],
        primary_key="concept_id",
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))

    match flavor:
        case "postgresql":
            assert lines == [
                'CREATE TABLE  "concept" (\n   "concept_id" BIGSERIAL PRIMARY KEY\n'
                '    , \n\n   "name" TEXT\n    NOT NULL\n);',
                'CREATE TABLE  "score" (\n   "concept" BIGINT PRIMARY KEY\n'
                '    CONSTRAINT fk_concept REFERENCES "concept"("concept_id") ON DELETE CASCADE\n'
                '    , \n\n   "score" INTEGER\n);',
                'CREATE UNIQUE INDEX concept_idx ON "concept" (\n  "name"\n);',
            ]
        case "sqlite":
            assert lines == [
                'CREATE TABLE  "concept" (\n   "concept_id" INTEGER PRIMARY KEY\n'
                '    , \n\n   "name" TEXT\n    NOT NULL\n);',
                'CREATE TABLE  "score" (\n   "concept" INTEGER PRIMARY KEY\n'
                '    CONSTRAINT fk_concept REFERENCES "concept"("concept_id") ON DELETE CASCADE\n'
                '    , \n\n   "score" INTEGER\n);',
                'CREATE UNIQUE INDEX concept_idx ON "concept" (\n  "name"\n);',
            ]
        case "mssql":
            assert lines == [
                'CREATE TABLE [concept] (\n'
                '   [concept_id] BIGINT IDENTITY(1,1) PRIMARY KEY\n'
                '    , \n\n'
                '   [name] NVARCHAR(200)\n'
                '    NOT NULL\n'
                ');',
                'CREATE TABLE [score] (\n'
                '   [concept] BIGINT PRIMARY KEY\n'
                '    CONSTRAINT fk_concept REFERENCES [concept]([concept_id]) ON DELETE '
                'CASCADE\n'
                '    , \n'
                '\n'
                '   [score] INT\n'
                ');',
                'CREATE UNIQUE INDEX [concept_idx] ON [concept] ([name]\n);']


def test_create_table_no_pk(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    Table(  # Regular table
        "concept",
        columns={
            "name": "varchar",
        },
        natural_key=["name"],
        schema=schema,
    )
    Table(  #  Table with no primary key and a fk in the nk
        "score",
        columns={
            "concept": "bigint",
            "score": "int",
        },
        natural_key=["concept"],
        primary_key=None,
        foreign_keys={
            "concept": "concept",
        },
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))
    (
        create_concept,
        create_score,
        add_score_concept_fk,
        create_concept_idx,
        create_score_idx,
    ) = map(strip_lines, lines)

    match flavor:
        case "postgresql":
            assert create_concept == [
                'CREATE TABLE  "concept" (',
                '"id" BIGSERIAL PRIMARY KEY',
                ",",
                '"name" TEXT',
                "NOT NULL",
                ");",
            ]
            assert add_score_concept_fk == [
                'ALTER TABLE "score"',
                'ADD COLUMN "concept" BIGINT NOT NULL',
                'CONSTRAINT fk_concept REFERENCES "concept"("id") ON DELETE CASCADE;',
            ]
        case "sqlite":
            assert create_concept == [
                'CREATE TABLE  "concept" (',
                '"id" INTEGER PRIMARY KEY',
                ",",
                '"name" TEXT',
                "NOT NULL",
                ");",
            ]
            assert add_score_concept_fk == [
                'ALTER TABLE "score"',
                'ADD COLUMN "concept" INTEGER NOT NULL',
                'CONSTRAINT fk_concept REFERENCES "concept"("id") ON DELETE CASCADE;',
            ]
        case "postgresql" | "sqlite":
            assert create_score == [
                'CREATE TABLE  "score" (',
                '"score" INTEGER',
                ");",
            ]
        case "mssql":
            assert create_concept == [
                "CREATE TABLE [concept] (",
                "[id] BIGINT IDENTITY(1,1) PRIMARY KEY",
                ",",
                "[name] NVARCHAR(200)",
                "NOT NULL",
                ");",
            ]
            assert create_score == [
                "CREATE TABLE [score] (",
                "[score] INT",
                ");",
            ]

    match flavor:
        case "postgresql" | "sqlite":
            assert create_concept_idx == [
                'CREATE UNIQUE INDEX concept_idx ON "concept" (',
                '"name"',
                ");",
            ]
            assert create_score_idx == [
                'CREATE UNIQUE INDEX score_idx ON "score" (',
                '"concept"',
                ");",
            ]
        case "mssql":
            assert create_concept_idx == [
                "CREATE UNIQUE INDEX [concept_idx] ON [concept] ([name]",
                ");",
            ]
            assert create_score_idx == [
                "CREATE UNIQUE INDEX [score_idx] ON [score] ([concept]",
                ");",
            ]


def test_create_table_no_pk_nk(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    Table(  #  Table with no primary key or natural key
        "score",
        columns={
            "score": "int",
        },
        natural_key=None,
        primary_key=None,
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))
    (create_score,) = map(strip_lines, lines)

    match flavor:
        case "postgresql" | "sqlite":
            assert create_score == [
                'CREATE TABLE  "score" (',
                '"score" INTEGER',
                ");",
            ]
        case "mssql":
            assert create_score == [
                "CREATE TABLE [score] (",
                "[score] INT",
                ");",
            ]


def test_create_unique_index():
    stmt = Statement("create_unique_index").table("my_table").natural_key(["name"])
    doc = stmt()
    lines = strip_lines(doc)
    assert lines == [
        'CREATE UNIQUE INDEX my_table_idx ON "my_table" (',
        '"name"',
        ");",
    ]



from typing import Optional


from nagra.schema import Schema
from nagra.table import Table
from nagra.exceptions import IncorrectSchema


class View:
    def __init__(
        self,
        name: str,
        view_columns: Optional[dict] = None,
        columns: Optional[dict] = None,
        natural_key: Optional[list[str]] = None,
        foreign_keys: Optional[dict] = None,
        as_select: Optional[str] = None,
        view_select: Optional[str] = None,
        view_where: Optional[str] = None,
        schema: Schema = Schema.default,
    ):
        self.name = name
        self.view_columns = view_columns
        self.as_select = as_select
        self.view_select = view_select
        self.schema = schema
        self.foreign_keys = foreign_keys or {}

        # Create underlying table
        if not columns:
            if not view_columns:
                msg = (
                    f"Error on view '{name}': one of "
                    "`columns` or `view_columns` must be defined"
                )
                raise IncorrectSchema(msg)

            columns = {}
            select = schema.get(self.view_select).select(*view_columns.values())
            dtypes = select.dtypes(with_optional=False)
            for col_name, (_, dt) in zip(view_columns, dtypes):
                columns[col_name] = dt.__name__

        self.table = Table(
            name,
            columns=columns,
            natural_key=natural_key,
            foreign_keys=foreign_keys,
            is_view=True,
            schema=self.schema,
        )
        # Add view to schema
        self.schema.add_view(self.name, self)

        # We try to quack like a table
        self.columns = self.table.columns

    def select(self, *columns, trn=None):
        return self.table.select(*columns, trn=trn)

    def view_def(self):
        if self.as_select:
            return self.as_select
        # generate a select from the referenced table
        stm = (
            self.schema.get(self.view_select)
            .select(*self.view_columns.values())
            .aliases(*self.view_columns.keys())
            .stm()
        )
        return stm.rstrip(";")

    @classmethod
    def get(self, name: str, schema: Schema = Schema.default):
        """
        Return view object for the given `name`
        """
        return schema.views.get(name)

    def eq(self, other):
        if not isinstance(other, View):
            return False
        if id(self) == id(other):
            return True
        ok = (
            self.name == other.name and
            self.view_def().strip() == other.view_def().strip()
        )
        return ok



import zoneinfo
from datetime import datetime, date
from uuid import UUID

import polars
import polars.datatypes
import polars.testing
import pytest


def test_to_polars(transaction, temperature):
    if transaction.flavor != "postgresql":
        pytest.skip("Sqlite not supported with polars")

    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    # Read data
    df = temperature.select().to_polars()
    columns = df.collect_schema().names()
    assert columns == ["timestamp", "city", "value"]
    # to_polars return a LazyFrame, we matertialize it
    df = df.collect()
    assert sorted(df["city"]) == ["Berlin", "London"]

    # Read with custom arg
    cond = "(= value {})"
    df = temperature.select().where(cond).to_polars(12).collect()
    assert list(df.columns) == ["timestamp", "city", "value"]
    assert sorted(df["city"]) == ["London"]


@pytest.mark.parametrize("batch", [True, False])
def test_from_polars(transaction, kitchensink, batch):
    if transaction.flavor != "postgresql":
        pytest.skip("Sqlite not supported with polars")

    json = {"a": 5, "b": [1, 2, 3]}

    df = polars.LazyFrame(
        {
            "varchar": ["ham"],
            "bigint": [1],
            "float": [1.0],
            "int": [1],
            "timestamp": ["1970-01-01 00:00:00"],
            "timestamptz": ["1970-01-01 00:00:00+00:00"],
            "bool": [True],
            "date": ["1970-01-01"],
            "json": [json],
            "uuid": ["F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"],
            "max": ["max"],
            "true": ["true"],
            "blob": [b"blob"],
        }
    )

    # UPSERT
    kitchensink.upsert().from_polars(df, batch=batch)
    (row,) = kitchensink.select()
    BRUTZ = zoneinfo.ZoneInfo(key="Europe/Brussels")
    assert row == (
        "ham",
        1,
        1.0,
        1,
        datetime(1970, 1, 1, 0, 0),
        datetime(1970, 1, 1, 1, 0, tzinfo=BRUTZ),
        True,
        date(1970, 1, 1),
        json,
        UUID("F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"),
        "max",
        "true",
        b"blob",
    )

    # SELECT with schema override for JSON
    json_df = (
        kitchensink.select("json")
        .to_polars(schema_overrides={"json": polars.datatypes.Object})
        .collect()
    )
    json_col = json_df["json"]
    assert json_col[0] == json

    # SELECT with operator and schema override
    new_df = (
        kitchensink.select(
            "(date_bin '5 days' timestamptz '1900-01-01')",
        )
        .aliases("ts")
        .to_polars(
            schema_overrides={
                "ts": polars.datatypes.Datetime(time_zone="Europe/Brussels")
            }
        )
        .collect()
    )

    ts = new_df["ts"]
    assert str(ts.dtype) == "Datetime(time_unit='us', time_zone='Europe/Brussels')"

    assert ts[0].isoformat() == "1969-12-30T01:00:00+01:00"


@pytest.mark.parametrize("batch", [True, False])
def test_with_chunking(transaction, temperature_no_nk_pk, batch):
    if transaction.flavor != "postgresql":
        pytest.skip("Sqlite not supported with polars")
    df = polars.DataFrame(
        {
            "timestamp": polars.datetime_range(
                datetime(1970, 1, 1), datetime(1970, 1, 2), "1s", eager=True
            ),
            "city": "City",
            "value": 20.0,
        }
    )
    temperature_no_nk_pk.insert("timestamp", "city", "value").from_polars(
        df.lazy(), batch=batch
    )
    result = temperature_no_nk_pk.select().to_polars().sort("timestamp").collect()

    polars.testing.assert_frame_equal(result, df)



from pathlib import Path

import pytest

from nagra import Table, Schema
from nagra.table import Column
from nagra.exceptions import IncorrectSchema
from nagra.transaction import Transaction


HERE = Path(__file__).parent


def test_toml_loader():
    # With a Path
    src = HERE / "assets" / "sample_schema.toml"
    test_schema = Schema.from_toml(src)
    table = Table.get("user", schema=test_schema)
    assert table is not None

    # with a string
    test_schema = Schema.from_toml(src.open().read())
    table = Table.get("user", schema=test_schema)
    assert table is not None

    # with an io base
    test_schema = Schema.from_toml(src.open())
    user_table = Table.get("user", schema=test_schema)
    blog_table = Table.get("blog", schema=test_schema)
    assert user_table is not None

    assert list(user_table.columns) == [
        "first_name",
        "last_name",
        "birthdate",
    ]
    assert list(blog_table.columns) == [
        "title",
        "length",
        "user",
    ]
    assert user_table.foreign_keys == {}
    assert blog_table.foreign_keys == {"user": "user"}
    assert user_table.primary_key == "id"
    assert blog_table.primary_key is None
    assert user_table.ctypes("postgresql", user_table.columns) == {
        "first_name": "TEXT",
        "last_name": "TEXT",
        "birthdate": "DATE",
    }
    assert blog_table.ctypes("postgresql", blog_table.columns) == {
        "title": "TEXT",
        "length": "INTEGER",
        "user": "BIGINT",
    }

    # Must fail when a duplicate table is added
    with pytest.raises(RuntimeError):
        Table(
            "user",
            columns={"name": "varchar"},
            natural_key=["name"],
            schema=test_schema,
        )

    # Test reset
    test_schema.reset()
    assert test_schema.tables == {}


def test_toml_generation():
    # Start from schema
    src = HERE / "assets" / "sample_schema.toml"
    test_schema = Schema.from_toml(src)

    clone_toml = test_schema.generate_toml()
    clone_schema = Schema()
    clone_schema.load_toml(clone_toml)
    assert test_schema.eq(clone_schema)


def test_bogus_fk(empty_transaction):
    with pytest.raises(IncorrectSchema):
        Table(
            "bogus",
            columns={
                "key": "uuid",
            },
            natural_key=["key"],
            foreign_keys={
                "key": "bogus",
            },
        )


def test_incorrect_nk(empty_transaction):
    with pytest.raises(IncorrectSchema):
        Table(
            "bad_nk",
            columns={
                "key": "uuid",
            },
            natural_key=["i_do_not_exist"],
        )


def test_create_tables(schema, empty_transaction):
    # Make sure we start from empty db
    assert not schema._db_columns(trn=empty_transaction)
    schema.create_tables(trn=empty_transaction)
    post = schema._db_columns(trn=empty_transaction)

    # Test person table is properly created
    assert "person" in post
    assert sorted(post["person"]) == ["id", "name", "parent"]

    # Add a column to existing table
    person = schema.tables["person"]
    person.columns["email"] = Column("email", "varchar")
    schema.create_tables(trn=empty_transaction)
    post = schema._db_columns(trn=empty_transaction)
    assert "person" in post
    assert sorted(post["person"]) == ["email", "id", "name", "parent"]

    # Needed to not polute other tests
    person.columns.pop("email")


def test_custom_id_type(empty_transaction):
    sch = Schema()
    city = Table(
        "city",
        columns={
            "id": "varchar",
            "name": "varchar",
        },
        schema=sch,
    )

    sch.drop()
    sch.create_tables()
    row_id = city.upsert("id", "name").execute("this-is-an-uuid", "test")
    assert row_id == "this-is-an-uuid"
    assert list(city.select()) == [("this-is-an-uuid", "test")]


def test_schema_from_nagra_db(transaction: Transaction):
    """
    Check introspection of a nagra created tables.
    """
    tables = [
        "address",
        "country",
        "kitchensink",
        "max_pop",
        "min_pop",
        "org",
        "parameter",
        "person",
        "population",
        "skill",
        "temperature",
        "temperature_no_nk_pk",
        "value",
    ]
    schema = Schema()
    schema.introspect_db()
    if transaction.flavor == "mssql":
        # We ignore table with array for mssql
        tables.remove("parameter")
    assert sorted(schema.tables) == tables
    assert all(schema.tables[n].is_view for n in ["max_pop", "min_pop"])

    views = [
        "max_pop",
        "min_pop",
    ]
    assert sorted(schema.views) == views

    # Check simple table
    person = schema.get("person")
    assert list(person.columns) == ["id", "name", "parent"]
    if transaction.flavor == "sqlite":
        expected = ["int", "str", "int"]
    else:
        expected = ["bigint", "str", "bigint"]

    assert [c.dtype for c in person.columns.values()] == expected
    assert person.foreign_keys == {"parent": "person"}
    assert person.primary_key == "id"
    assert person.natural_key == ["name"]

    # Check table with arrays
    if transaction.flavor == "mssql":
        pytest.skip("MSSQL does not support array columns")
    parameter = schema.get("parameter")
    assert list(parameter.columns) == ["id", "name", "timestamps", "values"]
    if transaction.flavor == "postgresql":
        expected_types = ["bigint", "str", "timestamp", "float"]
        assert [c.dtype for c in parameter.columns.values()] == expected_types
        expected_dims = ["", "", "[]", "[]"]
        assert [c.dims for c in parameter.columns.values()] == expected_dims
    else:
        expected_types = ["int", "str", "json", "json"]
        assert [c.dtype for c in parameter.columns.values()] == expected_types

    assert parameter.foreign_keys == {}
    assert parameter.primary_key == "id"
    assert parameter.natural_key == ["name"]


def test_schema_from_db(transaction: Transaction):
    """
    Check introspection on various coner cases
    """
    stmts = [
        "CREATE TABLE patient (patient_id int primary key)",
        # Primary key is also a foreign key
        """
        CREATE TABLE death (
           patient_id int primary key CONSTRAINT fk_patient REFERENCES patient(patient_id),
           death_date date
        )
        """,
        # One column of the natural key is a foreign key
        """
        CREATE TABLE visit (
           patient_id int primary key CONSTRAINT fk_visit_patient REFERENCES patient(patient_id),
           visit_date date
        )
        """,
        "CREATE UNIQUE INDEX visit_idx ON visit (patient_id, visit_date)",
    ]
    for stm in stmts:
        transaction.execute(stm)

    schema = Schema()
    schema.introspect_db()

    patient = schema.get("patient")
    assert list(patient.columns) == ["patient_id"]
    assert patient.primary_key == "patient_id"

    death = schema.get("death")
    assert list(death.columns) == ["patient_id", "death_date"]
    assert death.primary_key == "patient_id"
    assert death.foreign_keys == {"patient_id": "patient"}

    visit = schema.get("visit")
    assert list(visit.columns) == ["patient_id", "visit_date"]
    assert visit.primary_key == "patient_id"
    assert visit.natural_key == ["patient_id", "visit_date"]
    assert visit.foreign_keys == {"patient_id": "patient"}


def test_suspend_fk(transaction: Transaction):
    if transaction.flavor == "mssql":
        pytest.skip("Support for disabling foreign keys with mssql not implemented")

    # Skip sqlite
    is_sqlite = transaction.flavor == "sqlite"

    schema = Schema()
    whitelist = ["person", "skill"]
    schema.introspect_db(*whitelist)

    before = schema._db_fk(*whitelist)
    with schema.suspend_fk():
        with_suspend = schema._db_fk(*whitelist)
    after = schema._db_fk(*whitelist)

    if not is_sqlite:
        # With sqlite, the foreign keys constraints are still there
        # but not enforced
        assert len(with_suspend) == 0
    assert sorted(before) == ["person", "skill"]
    assert sorted(after) == ["person", "skill"]
    assert sorted(before["person"]) == ["fk_parent"]
    assert sorted(after["person"]) == ["fk_parent"]



import argparse
import os
from itertools import chain

from nagra import Transaction, Schema,  __version__
from nagra.utils import print_table, partition


def select(args, schema):
    table = schema.get(args.table)
    if args.columns:
        cols = args.columns
    else:
        # Ignore pk and blob col by default
        cols = table.default_columns(skip_pk=True, skip_blob=True)

    # Allow key=value syntax for positional args
    eq_conds, cols = partition(lambda c: "=" in c, cols)
    eq_where = []
    eq_args = []
    for cond in eq_conds:
        k, v = cond.split("=", 1)
        eq_where.append("(= %s {})" % k)
        eq_args.append(v)

    select = table.select(*cols)
    # Chain all where conditions (we allow multiple --where args)
    where = eq_where + args.where
    if where:
        select = select.where(*where)
    if args.limit:
        select = select.limit(args.limit)
    if args.orderby:
        orderby = chain.from_iterable(args.orderby)
        select = select.orderby(*orderby)
    rows = list(select.execute(*eq_args))
    headers = [d[0] for d in select.dtypes()]

    print_table(rows, headers, args.pivot, format=args.table_fmt)


def delete(args, schema):
    delete = schema.get(args.table).delete()
    # Allow key=value syntax for positional args
    eq_conds, cols = partition(lambda c: "=" in c, args.eq_cond)
    eq_where = []
    eq_args = []
    for cond in eq_conds:
        k, v = cond.split("=", 1)
        eq_where.append("(= %s {})" % k)
        eq_args.append(v)

    where = eq_where + args.where
    if where:
        delete = delete.where(*where)
    delete.execute(*eq_args)


def init(args, schema):
    schema.create_tables()


def print_schema(args, schema):
    if args.fmt == "d2":
        print(schema.generate_d2())
        return

    elif args.fmt == "toml":
        print(schema.generate_toml())
        return

    elif args.fmt == "pydantic":
        print(schema.generate_pydantic_models(table_names=args.tables))
        return

    # If tables name are given, print details
    if args.tables:
        rows = []
        headers = ["table", "column", "type"]
        for table_name in args.tables:
            for col in schema.get(table_name).columns.values():
                rows.append([table_name, col.name, col.dtype])
        print_table(rows, headers, args.pivot, format=args.table_fmt)
        return

    # List all tables
    rows = [(tbl.name, tbl.is_view) for tbl in schema.tables.values()]
    headers = ["table", "view"]
    print_table(sorted(rows), headers, args.pivot, format=args.table_fmt)


def show_version():
    print(__version__)


def run():
    # top-level parser
    parser = argparse.ArgumentParser(
        prog="nagra",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    default_db = os.environ.get("NAGRA_DB")
    default_schema = os.environ.get("NAGRA_SCHEMA")
    parser.add_argument(
        "--db",
        "-d",
        default=default_db,
        help=f"DB uri, (default: {default_db})",
    )
    parser.add_argument(
        "--schema",
        "-s",
        default=default_schema,
        help=f"DB schema, (default: {default_schema})",
    )
    parser.add_argument(
        "--pivot",
        "-p",
        action="store_true",
        help="Pivot results (one key-value table per record)",
    )
    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        help="Show version",
    )
    parser.add_argument(
        "--csv",
        action="store_const",
        const="csv",
        dest="table_fmt",
        help="Format output as csv",
    )
    subparsers = parser.add_subparsers(dest="command")

    parser_select = subparsers.add_parser("select")
    parser_select.add_argument("table")
    parser_select.add_argument("columns", nargs="*")
    parser_select.add_argument("--where", "-W", type=str, action="append", default=[])
    parser_select.add_argument("--limit", "-L", type=int)
    parser_select.add_argument(
        "--orderby",
        "-O",
        type=str,
        action="append",
        nargs="*",
        help="Order by given columns",
    )
    parser_select.set_defaults(func=select)

    parser_delete = subparsers.add_parser("delete")
    parser_delete.add_argument("table")
    parser_delete.add_argument("eq_cond", nargs="*", help="Simplified filters (eg: id=123)")
    parser_delete.add_argument("--where", "-W", type=str, action="append", default=[])
    parser_delete.set_defaults(func=delete)

    parser_init = subparsers.add_parser("init")
    parser_init.set_defaults(func=init)


    parser_schema = subparsers.add_parser("schema")
    parser_schema.add_argument(
        "--d2",
        help="Generate d2 file",
        dest="fmt",
        action="store_const",
        const="d2",
        default=None,
    )
    parser_schema.add_argument(
        "--toml",
        help="Generate toml file",
        dest="fmt",
        action="store_const",
        const="toml",
        default=None,
    )
    parser_schema.add_argument("tables", nargs="*")
    parser_schema.add_argument(
        "--pydantic",
        dest="fmt",
        action="store_const",
        const="pydantic",
        default=None,
        help="Generate pydantic models for the schema",
    )
    parser_schema.set_defaults(func=print_schema)

    # Parse args
    args = parser.parse_args()
    if args.version:
        show_version()
        return
    if not args.command:
        parser.print_help()
        return

    try:
        with Transaction(args.db):
            if args.schema:
                schema = Schema.from_toml(open(args.schema))
            else:
                schema = Schema.from_db()
            args.func(args, schema=schema)
    except (BrokenPipeError, KeyboardInterrupt):
        pass



import sqlite3
import threading
from itertools import islice
from typing import Callable

from nagra.utils import logger, UNSET, mssql_connection_string
from nagra.exceptions import NoActiveTransaction, TransactionReenterError


class LRUGenerator:
    """
    LRUGenerator wraps generator calls and implement LRU caching
    """

    def __init__(self, generator, size=1000):
        self.size = size
        self.generator = generator
        self.recent = {}
        self.older = {}

    def run(self, keys):
        # Capture existing known values in `cached`
        cached = {}
        fresh = []
        for k in keys:
            v = self.get(k)
            if v is UNSET:
                fresh.append(k)
            else:
                cached[k] = v

        # Update internal state with generator results
        for fresh_k, fresh_v in zip(fresh, self.generator(fresh)):
            self.set(fresh_k, fresh_v)

        # Yield results
        for k in keys:
            if k in cached:
                yield cached[k]
            else:
                yield self.get(k)

        # Vaccum
        self.vaccum()

    def __contains__(self, key):
        return key in self.recent or key in self.older

    def get(self, key, default=UNSET):
        if key in self.recent:
            return self.recent[key]

        if key in self.older:
            value = self.older[key]
            self.recent[key] = value
            return value

        return default

    def set(self, key, value):
        self.recent[key] = value

    def update(self, values):
        self.recent.update(values)
        self.vaccum()

    def vaccum(self):
        if len(self.older) + len(self.recent) > self.size:
            self.older = self.recent
            self.recent = {}


class Transaction:
    _stack_lock = threading.Lock()
    # _local_stack: ContextVar[list["Transaction"]] = ContextVar('_local_stack', default=[])
    _local = threading.local()
    _local.stack = []

    def __init__(self, dsn, rollback=False, fk_cache=False):
        self.auto_rollback = rollback
        self._fk_cache = {} if fk_cache else None

        if dsn.startswith("postgresql://"):
            try:
                import psycopg
            except ImportError as exc:  # pragma: no cover - optional dependency
                msg = "Postgresql support requires the 'psycopg' package. Install nagra[pg]."
                raise ImportError(msg) from exc

            # TODO use Connection Pool
            self.flavor = "postgresql"
            self.connection = psycopg.connect(dsn)
        elif dsn.startswith("sqlite://"):
            self.flavor = "sqlite"
            filename = dsn[9:]
            self.connection = sqlite3.connect(filename)
            self.connection.execute("PRAGMA foreign_keys = 1")
        elif dsn.startswith("mssql://"):
            try:
                import pyodbc
            except ImportError as exc:  # pragma: no cover - optional dependency
                msg = "SQL Server support requires the 'pyodbc' package. Install nagra[mssql]."
                raise ImportError(msg) from exc

            self.flavor = "mssql"
            conn_str = mssql_connection_string(dsn)
            self.connection = pyodbc.connect(conn_str, autocommit=False)
            cursor = self.connection.cursor()
            cursor.execute("SET QUOTED_IDENTIFIER ON")
            cursor.execute("SET XACT_ABORT ON")  # Enforce atomicity
            cursor.close()
        elif dsn.startswith("duckdb://"):
            import duckdb

            self.flavor = "duckdb"
            filename = dsn[9:]
            self.connection = duckdb.connect(filename)
            self.connection.begin()
        else:
            raise ValueError(f"Invalid dsn string: {dsn}")

    def execute(self, stmt, args=tuple()) -> "ResultCursor":
        logger.debug(stmt)
        cursor = self.connection.cursor()
        cursor.execute(stmt, args)
        match self.flavor:
            case "postgresql" | "sqlite":
                return ResultCursor(cursor)
            case "mssql":
                return MSSQLCursor(cursor)
            case _:
                msg = f"Unsupported flavor for execute: {self.flavor}"
                raise RuntimeError(msg)

    def executemany(
        self, stmt, args=None, returning=False
    ) -> "ResultCursor | RowCursor":
        logger.debug(stmt)
        cursor = self.connection.cursor()
        args = args or []

        match self.flavor:
            case "postgresql":
                cursor.executemany(stmt, args, returning=returning)
                return ResultCursor(cursor, returning=returning)
            case "sqlite":
                cursor.executemany(stmt, args)
                return ResultCursor(cursor)
            case "mssql":
                if not returning:
                    cursor.fast_executemany = True
                    cursor.executemany(stmt, args)
                    return ResultCursor(cursor, returning=returning)
                else:
                    rows = self._executemany_mssql(cursor, stmt, args)
                    return RowCursor(rows)
            case _:
                msg = f"Unsupported flavor for executemany: {self.flavor}"
                raise RuntimeError(msg)

    def _executemany_mssql(self, cursor, stmt, args):
        import pyodbc

        for params in args:
            cursor.execute(stmt, params)
            try:
                row = cursor.fetchone()
            except pyodbc.Error:
                row = None
            yield row

    def rollback(self):
        self.connection.rollback()

    def commit(self):
        self.connection.commit()

    def __enter__(self):
        Transaction.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        Transaction.pop(self)
        if self.auto_rollback or exc_type is not None:
            self.rollback()
        else:
            self.commit()
        self.close()

    def close(self):
        self.connection.close()

    @classmethod
    def push(cls, transaction):
        with cls._stack_lock:
            if not hasattr(cls._local, "stack"):
                cls._local.stack = []
            if transaction in cls._local.stack:
                raise TransactionReenterError(
                    "Transaction already in stack. Are you entering a context with the same transaction twice?"
                )
            cls._local.stack.append(transaction)
            # stack = cls._local_stack.get()
            # stack.append(transaction)
            # cls._local_stack.set(stack)

    @classmethod
    def pop(cls, expected_trn):
        with cls._stack_lock:
            trn = cls._local.stack.pop()
            assert trn is expected_trn, "Unexpected Transaction when leaving context"

            # stack = cls._local_stack.get()
            # trn = stack.pop()
            # cls._local_stack.set(stack)
            # assert id(trn) == id(expected_trn), "Unexpected Transaction when leaving context"

    @classmethod
    def current(cls) -> "Transaction | DummyTransaction":
        try:
            with cls._stack_lock:
                return cls._local.stack[-1]
                # return cls._local_stack.get()[-1]
        except (IndexError, AttributeError):
            return dummy_transaction

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.flavor}>"

    def get_fk_cache(
        self, cache_key: tuple[str, ...], fn: Callable
    ) -> LRUGenerator | None:
        """
        Instanciate an LRUGenerator for the given function
        `fn`. Use `cache_key` to identify them. Will return
        `None` if `fk_cache` is False.
        """
        if self._fk_cache is None:
            return None

        if lru := self._fk_cache.get(cache_key):
            return lru
        lru = LRUGenerator(fn)
        self._fk_cache[cache_key] = lru
        return lru


def yield_from_cursor(cursor):
    while rows := cursor.fetchmany(1000):
        yield from rows


class CursorMixin:
    """
    Provide extra properties (one, all, scalar, scalars) and
    usefull methods for Cursor classes.
    """

    def fetchone(self):
        return next(self, None)

    def fetchmany(self, size=1000):
        return list(islice(self, size))

    def fetchall(self):
        return list(self)

    @property
    def one(self):
        return self.fetchone()

    @property
    def all(self):
        return self.fetchall()

    @property
    def scalar(self):
        (res,) = self.fetchone()
        return res

    @property
    def scalars(self):
        for (res,) in self:
            yield res


class ResultCursor(CursorMixin):
    def __init__(self, native_cursor, returning=False):
        self.native_cursor = native_cursor
        self.returning = returning

    def __iter__(self):
        if not self.returning:
            return iter(self.native_cursor)
        return self.iter_returning()

    def iter_returning(self):
        # Insert/update queries returning data must be iterated in a
        # different fashion
        while True:
            row = self.native_cursor.fetchone()
            yield row
            if not self.native_cursor.nextset():
                break

    def __next__(self):
        return next(iter(self))

    def close(self):
        self.native_cursor.close()


class MSSQLCursor(ResultCursor):
    def __iter__(self):
        return (r and tuple(r) for r in super().__iter__())

    def __next__(self):
        return next(self.native_cursor)


class RowCursor:
    """
    Wrapper around a collection of rows that mimicks ResultCursor,
    needed for mssql support.
    """

    def __init__(self, rows):
        self.rows = iter(rows)

    def __iter__(self):
        yield from (r and tuple(r) for r in self.rows)


class ExecMany:
    """
    Helper class that can consume an iterator and feed the values
    yielded to a (returning) executemany statement.
    """

    def __init__(self, stm, values, trn):
        self.stm = stm
        self.values = values
        self.trn = trn

    def __iter__(self):
        # Use a dedicated cursor to allow concurrent execution
        logger.debug(self.stm)
        match self.trn.flavor:
            case "sqlite":
                cursor = self.trn.connection.cursor()
                for vals in self.values:
                    cursor.execute(self.stm, vals)
                    yield cursor.fetchone()
            case "postgresql" | "mssql":
                cursor = self.trn.executemany(
                    self.stm,
                    self.values,
                    returning=True,
                )
                yield from cursor


class DummyTransaction(Transaction):
    """
    Postgresql flavored transaction look-alike
    """

    flavor = "postgresql"

    def __init__(self):
        pass

    def execute(self, stmt, args=tuple()):
        raise NoActiveTransaction()

    def executemany(self, stmt, args=None, returning=True):
        raise NoActiveTransaction()


dummy_transaction = DummyTransaction()
